import logging
from io import StringIO
from typing import Dict
from logilab.common.graph import ordered_nodes

import yams.types as yams_types


def serialize_to_python(schema: yams_types.Schema) -> str:
    out = StringIO()
    w = out.write
    w("from yams.buildobjs import *\n\n")

    graph: Dict = {}

    for entity in schema.entities():
        graph.setdefault(entity, [])

    for e in reversed(ordered_nodes(graph)):
        if not e.final:
            if e._specialized_type:
                base = e._specialized_type
            else:
                base = "EntityType"
            w("class %s(%s):\n" % (e.type, base))
            attr_defs = list(e.attribute_definitions())
            if attr_defs:
                for attr, obj in attr_defs:
                    w("    %s = %s()\n" % (attr.type, obj.type))
            else:
                w("    pass\n")
            w("\n")

    for r in schema.relations():
        if not r.final:
            if r.subjects() and r.objects():
                w("class %s(RelationDefinition):\n" % r.type)
                w("    subject = (%s,)\n" % ", ".join("'%s'" % x for x in r.subjects()))
                w("    object = (%s,)\n" % ", ".join("'%s'" % x for x in r.objects()))
                w("\n")
            else:
                logging.warning("relation definition %s missing subject/object" % r.type)

    return out.getvalue()



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""ER schema loader.

Use either a sql derivated language for entities and relation definitions
files or a direct python definition file.
"""

import sys
import os
import types
import pkgutil
import _frozen_importlib_external
from os import listdir
from os.path import dirname, exists, join, splitext, basename, abspath, realpath
from warnings import warn
from typing import Any, List, Tuple, Dict, Callable, Sequence, Optional, Type, cast

from logilab.common import tempattr, deprecation
from logilab.common.modutils import modpath_from_file, cleanup_sys_modules, clean_sys_modules

from yams import BadSchemaDefinition
from yams import constraints, schema as schemamod
from yams import buildobjs

import yams.types as yams_types

__docformat__: str = "restructuredtext en"

CONSTRAINTS: Dict[str, Callable[..., constraints.BaseConstraint]] = {}

# add constraint classes to the context
for object_name in dir(constraints):
    if object_name[0] == "_":
        continue

    object_ = getattr(constraints, object_name)

    try:
        if issubclass(object_, constraints.BaseConstraint) and (
            object_ is not constraints.BaseConstraint
        ):
            CONSTRAINTS[object_name] = object_
    except TypeError:
        continue


@deprecation.argument_renamed(old_name="erdefs", new_name="entity_relation_definitions")
@deprecation.argument_renamed(
    old_name="remove_unused_rtypes", new_name="remove_unused_relation_types"
)
def fill_schema(
    schema: yams_types.Schema,
    entity_relation_definitions: Dict,
    register_base_types: bool = True,
    remove_unused_relation_types: bool = False,
    post_build_callbacks: List[Callable[[Any], Any]] = [],
) -> yams_types.Schema:

    if register_base_types:
        buildobjs.register_base_types(schema)

    # relation definitions may appear multiple times
    entity_relation_definitions_values: set = set(entity_relation_definitions.values())

    # register relation types and non final entity types
    for definition in entity_relation_definitions_values:
        if isinstance(definition, type):
            definition = definition()

        if isinstance(definition, buildobjs.RelationType):
            schema.add_relation_type(definition)

        elif isinstance(definition, buildobjs.EntityType):
            schema.add_entity_type(definition)

    # register relation definitions
    for definition in entity_relation_definitions_values:
        if isinstance(definition, type):
            definition = definition()

        definition.expand_relation_definitions(entity_relation_definitions, schema)

    # call 'post_build_callback' functions found in schema modules
    for callback in post_build_callbacks:
        callback(schema)

    # finalize schema
    schema.finalize()

    # check permissions are valid on entities and relations
    for entities_and_relations_schema in schema.entities() + schema.relations():  # type: ignore
        entities_and_relations_schema.check_permission_definitions()

    # check unique together consistency
    for entity_schema in schema.entities():
        entity_schema.check_unique_together()

    # optionaly remove relation types without definitions
    if remove_unused_relation_types:
        for relation_schema in schema.relations():
            if not relation_schema.relation_definitions:
                schema.del_relation_type(relation_schema)

    return schema


@deprecation.attribute_renamed(old_name="schemacls", new_name="schema_class")
@deprecation.attribute_renamed(old_name="extrapath", new_name="extra_path")
class SchemaLoader:
    """the schema loader is responsible to build a schema object from a
    set of files
    """

    schema_class: Type[schemamod.Schema] = schemamod.Schema
    extra_path: Optional[str] = None
    context: Dict[str, Callable] = dict(
        [(attr, getattr(buildobjs, attr)) for attr in buildobjs.__all__]
    )
    context.update(CONSTRAINTS)

    @deprecation.argument_renamed(
        old_name="remove_unused_rtypes", new_name="remove_unused_relation_types"
    )
    @deprecation.argument_renamed(old_name="modnames", new_name="module_names")
    def load(
        self,
        module_names: Sequence[Tuple[Any, str]],
        name: Optional[str] = None,
        register_base_types: bool = True,
        construction_mode: str = "strict",
        remove_unused_relation_types: bool = True,
    ) -> yams_types.Schema:
        """return a schema from the schema definition read from <module_names> (a
        list of (PACKAGE, module_name))
        """

        self.defined: Dict = {}
        self.loaded_files: List = []
        self.post_build_callbacks: List = []
        # mypy: Module has no attribute "context"
        sys.modules[__name__].context = self  # type: ignore

        # ensure we don't have an iterator
        module_names = tuple(module_names)

        # legacy usage using a directory list
        is_directories = module_names and not isinstance(module_names[0], (list, tuple))
        try:
            if is_directories:
                warn("provide a list of modules names instead of directories", DeprecationWarning)
                self._load_definition_files(module_names)  # type: ignore # retrocompat situation

            else:
                self._load_module_names(module_names)

            schema = self.schema_class(name or "NoName", construction_mode=construction_mode)
            # if construction_mode != "strict" handle errors

            try:
                fill_schema(
                    schema,
                    self.defined,
                    register_base_types,
                    remove_unused_relation_types=remove_unused_relation_types,
                    post_build_callbacks=self.post_build_callbacks,
                )
            except Exception as exception:
                if not hasattr(exception, "schema_files"):
                    # mypy: "Exception" has no attribute "schema_files"
                    # XXX looks like a hack to transport information
                    exception.schema_files = self.loaded_files  # type: ignore

                raise
        finally:
            # cleanup sys.modules from schema modules
            # ensure we're only cleaning schema [sub]modules
            if is_directories:
                directories = [
                    (
                        not directory.endswith(  # type: ignore
                            os.sep + self.main_schema_directory  # type: ignore # retrocompat
                        )
                        and join(directory, self.main_schema_directory)  # type: ignore
                        or directory
                    )
                    for directory in module_names
                ]

                cleanup_sys_modules(directories)

            else:
                clean_sys_modules([mname for _, mname in module_names])

        # mypy: "Schema" has no attribute "loaded_files"
        # another dynamic attribute
        schema.loaded_files = self.loaded_files  # type: ignore

        return schema

    def _load_definition_files(self, directories: Sequence[str]) -> None:
        for directory in directories:
            package = basename(directory)

            for file_path in self.get_schema_files(directory):
                with tempattr(buildobjs, "PACKAGE", package):
                    self.handle_file(file_path, None)

    @deprecation.argument_renamed(old_name="modnames", new_name="module_names")
    def _load_module_names(self, module_names: Sequence[Tuple[Any, str]]) -> None:
        for package, module_name in module_names:
            loader = pkgutil.find_loader(module_name)

            if loader is None:
                continue

            assert isinstance(loader, _frozen_importlib_external.FileLoader)
            file_path = loader.get_filename()

            if file_path.endswith(".pyc"):
                # check that related source file exists and ensure passing a
                # .py file to exec_file()
                file_path = file_path[:-1]

                if not exists(file_path):
                    continue

            with tempattr(buildobjs, "PACKAGE", package):
                self.handle_file(file_path, module_name=module_name)

    _load_modnames = deprecation.renamed(old_name="_load_modnames", new_function=_load_module_names)

    # has to be overridable sometimes (usually for test purpose)
    main_schema_directory: str = "schema"

    def get_schema_files(self, directory: str) -> List[str]:
        """return an ordered list of files defining a schema

        look for a schema.py file and or a schema sub-directory in the given
        directory
        """
        result = []

        if exists(join(directory, self.main_schema_directory + ".py")):
            result = [join(directory, self.main_schema_directory + ".py")]

        if exists(join(directory, self.main_schema_directory)):
            directory = join(directory, self.main_schema_directory)

            for filename in listdir(directory):
                if filename[0] == "_":
                    if filename == "__init__.py":
                        result.insert(0, join(directory, filename))

                    continue

                extension = splitext(filename)[1]

                if extension == ".py":
                    result.append(join(directory, filename))
                else:
                    self.unhandled_file(join(directory, filename))

        return result

    @deprecation.argument_renamed(old_name="filepath", new_name="file_path")
    @deprecation.argument_renamed(old_name="modname", new_name="module_name")
    def handle_file(self, file_path: str, module_name: Optional[str] = None) -> None:
        """handle a partial schema definition file according to its extension"""
        assert file_path.endswith(".py"), "not a python file"

        if file_path not in self.loaded_files:
            module_name, module = self.exec_file(file_path, module_name)
            objects_to_add = set()

            for name, object_ in vars(module).items():
                if (
                    isinstance(object_, type)
                    and issubclass(object_, buildobjs.Definition)
                    and object_.__module__ == module_name
                    and not name.startswith("_")
                ):
                    objects_to_add.add(object_)

            for object_ in objects_to_add:
                self.add_definition(object_, file_path)

            if hasattr(module, "post_build_callback"):
                # mypy: Module has no attribute "post_build_callback"
                # it is tested just before in the if
                self.post_build_callbacks.append(module.post_build_callback)  # type: ignore

            self.loaded_files.append(file_path)

    @deprecation.argument_renamed(old_name="filepath", new_name="file_path")
    def unhandled_file(self, file_path: str) -> None:
        """called when a file without handler associated has been found,
        does nothing by default.
        """

    @deprecation.argument_renamed(old_name="filepath", new_name="file_path")
    @deprecation.argument_renamed(old_name="defobject", new_name="definition_object")
    def add_definition(
        self, definition_object: Type[buildobjs.Definition], file_path: Optional[str] = None
    ) -> None:
        """file handler callback to add a definition object

        wildcard capability force to load schema in two steps : first register
        all definition objects (here), then create actual schema objects (done in
        `_build_schema`)
        """
        if not issubclass(definition_object, buildobjs.Definition):
            raise BadSchemaDefinition(file_path, "invalid definition object")

        definition_object.expand_type_definitions(self.defined)

    @deprecation.argument_renamed(old_name="filepath", new_name="file_path")
    @deprecation.argument_renamed(old_name="modname", new_name="module_name")
    def exec_file(self, file_path: str, module_name: Optional[str]) -> Tuple[str, types.ModuleType]:
        if module_name is None:
            try:
                module_name = ".".join(modpath_from_file(file_path, self.extra_path))
            except ImportError:
                warn(
                    "module for %s can't be found, add necessary __init__.py "
                    "files to make it importable" % file_path,
                    DeprecationWarning,
                )

                module_name = splitext(basename(file_path))[0]

        cast(str, module_name)

        if module_name in sys.modules:
            module: types.ModuleType = sys.modules[module_name]

            # NOTE: don't test raw equality to avoid .pyc / .py comparisons
            mpath: str = realpath(abspath(module.__file__))
            fpath: str = realpath(abspath(file_path))

            assert mpath.startswith(fpath), (module_name, file_path, module.__file__)

        else:
            file_globals: Dict[str, str] = {}  # self.context.copy()
            file_globals["__file__"] = file_path
            file_globals["__name__"] = module_name

            package: str = ".".join(module_name.split(".")[:-1])

            if package and package not in sys.modules:
                __import__(package)

            with open(file_path) as f:
                try:
                    code = compile(f.read(), file_path, "exec")
                    exec(code, file_globals)
                except Exception:
                    print("exception while reading %s" % file_path, file=sys.stderr)
                    raise

            file_globals["__file__"] = file_path

            module = types.ModuleType(str(module_name))
            module.__dict__.update(file_globals)

            sys.modules[module_name] = module

            if package:
                setattr(sys.modules[package], module_name.split(".")[-1], module)

            if basename(file_path) == "__init__.py":
                # add __path__ to make dynamic loading work as defined in PEP 302
                # https://www.python.org/dev/peps/pep-0302/#packages-and-the-role-of-path
                module.__path__ = [dirname(file_path)]  # type: ignore # dynamic attribute

        return (module_name, module)


def fill_schema_from_namespace(
    schema: yams_types.Schema, items: Sequence[Tuple[Any, Any]], **kwargs
) -> None:
    entity_relation_definitions: Dict = {}

    for _, object_ in items:
        if (
            isinstance(object_, type)
            and issubclass(object_, buildobjs.Definition)
            and object_
            not in (buildobjs.Definition, buildobjs.RelationDefinition, buildobjs.EntityType)
        ):
            object_.expand_type_definitions(entity_relation_definitions)

    fill_schema(schema, entity_relation_definitions, **kwargs)


def build_schema_from_namespace(items: Sequence[Tuple[Any, Any]]) -> schemamod.Schema:
    schema = schemamod.Schema("noname")

    fill_schema_from_namespace(schema, items)

    return schema


class _Context:
    def __init__(self) -> None:
        self.defined: Dict = {}


context: _Context = _Context()



# copyright 2004-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.

from logilab.common.testlib import mock_object

import unittest

from yams.constraints import (
    Attribute,
    BoundaryConstraint,
    FormatConstraint,
    IntervalBoundConstraint,
    NOW,
    RegexpConstraint,
    SizeConstraint,
    StaticVocabularyConstraint,
    TODAY,
    UniqueConstraint,
)
from datetime import datetime, date, timedelta


class ConstraintTC(unittest.TestCase):

    if not hasattr(unittest.TestCase, "subTest"):
        from contextlib import contextmanager

        @contextmanager
        def subTest(self, **kwargs):
            yield

    def test_membership(self):
        s = set()
        cstrs = [
            UniqueConstraint(),
            SizeConstraint(min=0, max=42),
            RegexpConstraint("babar", 0),
            BoundaryConstraint(">", 1),
            IntervalBoundConstraint(minvalue=0, maxvalue=42),
            StaticVocabularyConstraint((1, 2, 3)),
            FormatConstraint(),
        ]
        for cstr in cstrs:
            s.add(cstr)
            s.add(type(cstr).deserialize(cstr.serialize()))
        self.assertEqual(7, len(s))

    def test_interval_serialization_integers(self):
        cstr = IntervalBoundConstraint(12, 13)
        self.assertEqual(IntervalBoundConstraint.deserialize("12;13"), cstr)
        self.assertEqual(cstr.serialize(), '{"maxvalue": 13, "minvalue": 12, "msg": null}')
        self.assertEqual(cstr.__class__.deserialize(cstr.serialize()), cstr)
        cstr = IntervalBoundConstraint(maxvalue=13)
        self.assertEqual(IntervalBoundConstraint.deserialize("None;13"), cstr)
        self.assertEqual(cstr.serialize(), '{"maxvalue": 13, "minvalue": null, "msg": null}')
        self.assertEqual(cstr.__class__.deserialize(cstr.serialize()), cstr)
        cstr = IntervalBoundConstraint(minvalue=13)
        self.assertEqual(IntervalBoundConstraint.deserialize("13;None"), cstr)
        self.assertEqual(cstr.serialize(), '{"maxvalue": null, "minvalue": 13, "msg": null}')
        self.assertEqual(cstr.__class__.deserialize(cstr.serialize()), cstr)
        self.assertRaises(AssertionError, IntervalBoundConstraint)

    def test_interval_serialization_floats(self):
        cstr = IntervalBoundConstraint(12.13, 13.14)
        self.assertEqual(IntervalBoundConstraint.deserialize("12.13;13.14"), cstr)
        self.assertEqual(cstr.serialize(), '{"maxvalue": 13.14, "minvalue": 12.13, "msg": null}')
        self.assertEqual(cstr.__class__.deserialize(cstr.serialize()), cstr)

    def test_interval_deserialization_integers(self):
        cstr = IntervalBoundConstraint.deserialize("12;13")
        self.assertEqual(cstr.minvalue, 12)
        self.assertEqual(cstr.maxvalue, 13)
        cstr = IntervalBoundConstraint.deserialize("None;13")
        self.assertEqual(cstr.minvalue, None)
        self.assertEqual(cstr.maxvalue, 13)
        cstr = IntervalBoundConstraint.deserialize("12;None")
        self.assertEqual(cstr.minvalue, 12)
        self.assertEqual(cstr.maxvalue, None)

    def test_interval_deserialization_floats(self):
        cstr = IntervalBoundConstraint.deserialize("12.13;13.14")
        self.assertEqual(cstr.minvalue, 12.13)
        self.assertEqual(cstr.maxvalue, 13.14)

    def test_interval_attribute_error(self):
        cstr = IntervalBoundConstraint(minvalue=Attribute("hip"), maxvalue=Attribute("hop"))

        class entity:
            hip, hop = 34, 42

        self.assertEqual(
            cstr.failed_message("key", 20, entity),
            (
                "value %(KEY-value)s must be >= %(KEY-boundary)s",
                {"key-boundary": "hip", "key-value": 20},
            ),
        )
        self.assertEqual(
            cstr.failed_message("key", 43, entity),
            (
                "value %(KEY-value)s must be <= %(KEY-boundary)s",
                {"key-boundary": "hop", "key-value": 43},
            ),
        )

    def test_regexp_serialization(self):
        cstr = RegexpConstraint("[a-z]+,[A-Z]+", 40)
        self.assertEqual(cstr.serialize(), '{"flags": 40, "msg": null, "regexp": "[a-z]+,[A-Z]+"}')
        self.assertEqual(cstr.__class__.deserialize(cstr.serialize()), cstr)

    def test_regexp_deserialization(self):
        cstr = RegexpConstraint.deserialize("[a-z]+,[A-Z]+,40")
        self.assertEqual(cstr.regexp, "[a-z]+,[A-Z]+")
        self.assertEqual(cstr.flags, 40)

    def test_interval_with_attribute(self):
        cstr = IntervalBoundConstraint(NOW(), Attribute("hop"))
        cstr2 = IntervalBoundConstraint.deserialize(cstr.serialize())
        self.assertEqual(cstr2.minvalue.offset, None)
        self.assertEqual(cstr2.maxvalue.attr, "hop")
        self.assertTrue(
            cstr2.check(
                mock_object(hop=datetime.now() + timedelta(hours=1)),
                "hip",
                datetime.now() + timedelta(seconds=2),
            )
        )
        # fail, value < minvalue
        self.assertFalse(
            cstr2.check(
                mock_object(hop=datetime.now() + timedelta(hours=1)),
                "hip",
                datetime.now() - timedelta(hours=2),
            )
        )
        # fail, value > maxvalue
        self.assertFalse(
            cstr2.check(
                mock_object(hop=datetime.now() + timedelta(hours=1)),
                "hip",
                datetime.now() + timedelta(hours=2),
            )
        )

    def test_interval_with_date(self):
        cstr = IntervalBoundConstraint(TODAY(timedelta(1)), TODAY(timedelta(3)))
        cstr2 = IntervalBoundConstraint.deserialize(cstr.serialize())
        self.assertEqual(cstr2.minvalue.offset, timedelta(1))
        self.assertEqual(cstr2.maxvalue.offset, timedelta(3))
        self.assertTrue(cstr2.check(None, "hip", date.today() + timedelta(2)))
        # fail, value < minvalue
        self.assertFalse(cstr2.check(None, "hip", date.today()))
        # fail, value > maxvalue
        self.assertFalse(cstr2.check(None, "hip", date.today() + timedelta(4)))

    def test_bound_constant(self):
        cstr = BoundaryConstraint("<=", 0)
        cstr2 = BoundaryConstraint.deserialize(cstr.serialize())
        self.assertFalse(cstr2.check(None, "hip", 25))
        self.assertTrue(cstr2.check(None, "hip", -1))

    def test_bound_with_attribute(self):
        cstr = BoundaryConstraint("<=", Attribute("hop"))
        cstr2 = BoundaryConstraint.deserialize(cstr.serialize())
        self.assertEqual(cstr, cstr2)
        self.assertEqual(cstr2.boundary.attr, "hop")
        self.assertEqual(cstr2.operator, "<=")
        self.assertTrue(cstr2.check(mock_object(hop=date.today()), "hip", date.today()))
        # fail, value > maxvalue
        self.assertFalse(
            cstr2.check(mock_object(hop=date.today()), "hip", date.today() + timedelta(days=1))
        )

    def test_bound_with_date(self):
        cstr = BoundaryConstraint("<=", TODAY())
        cstr2 = BoundaryConstraint.deserialize(cstr.serialize())
        self.assertEqual(cstr, cstr2)
        self.assertEqual(cstr2.boundary.offset, None)
        self.assertEqual(cstr2.operator, "<=")
        self.assertTrue(cstr2.check(None, "hip", date.today()))
        # fail, value > maxvalue
        self.assertFalse(cstr2.check(None, "hip", date.today() + timedelta(days=1)))

    def test_bound_with_unset_attribute(self):
        cstr = BoundaryConstraint("<=", None)
        self.assertTrue(cstr.check(None, "hip", date.today()))
        cstr = BoundaryConstraint("<=", Attribute("unset_attr"))
        self.assertTrue(cstr.check(mock_object(unset_attr=None), "hip", date.today()))

    def test_boundary_constraint_message(self):
        cstr = BoundaryConstraint("<=", 0)
        self.assertEqual(
            cstr.failed_message("attr", 1, object()),
            (
                "value %(KEY-value)s must be <= %(KEY-boundary)s",
                {"attr-value": 1, "attr-boundary": 0},
            ),
        )

    def test_vocab_constraint_serialization(self):
        cstr = StaticVocabularyConstraint(["a, b", "c"])
        self.assertEqual(
            StaticVocabularyConstraint.deserialize(cstr.serialize()).values, ("a, b", "c")
        )

    def test_custom_message(self):
        cstrs = [
            UniqueConstraint(msg="constraint failed, you monkey!"),
            SizeConstraint(min=0, max=42, msg="constraint failed, you monkey!"),
            RegexpConstraint("babar", 0, msg="constraint failed, you monkey!"),
            BoundaryConstraint(">", 1, msg="constraint failed, you monkey!"),
            IntervalBoundConstraint(minvalue=0, maxvalue=42, msg="constraint failed, you monkey!"),
            StaticVocabularyConstraint((1, 2, 3), msg="constraint failed, you monkey!"),
            FormatConstraint(msg="constraint failed, you monkey!"),
        ]
        for cstr in cstrs:
            with self.subTest(cstr=cstr):
                self.assertEqual(
                    cstr.failed_message("key", "value", object()),
                    ("constraint failed, you monkey!", {}),
                )
            cstr = type(cstr).deserialize(cstr.serialize())
            with self.subTest(cstr=cstr):
                self.assertEqual(
                    cstr.failed_message("key", "value", object()),
                    ("constraint failed, you monkey!", {}),
                )


if __name__ == "__main__":
    unittest.main()



from logilab.common import testlib

from yams.buildobjs import register_base_types, EntityType, RelationType, RelationDefinition
from yams import schema, serialize


class MyTests(testlib.TestCase):
    def test_yams_serialize(self):
        s = schema.Schema("Unknown")
        register_base_types(s)
        e2 = EntityType("Ent2")
        s.add_entity_type(e2)
        e1 = EntityType("Ent1")
        e1.specialized_type = "Ent2"
        s.add_entity_type(e1)
        s.add_relation_type(RelationType("attr1"))
        s.add_relation_def(RelationDefinition("Ent1", "attr1", "String"))
        out = serialize.serialize_to_python(s)
        self.assertMultiLineEqual(
            out,
            "\n".join(
                [
                    "from yams.buildobjs import *",
                    "",
                    "class Ent2(EntityType):",
                    "    pass",
                    "",
                    "class Ent1(Ent2):",
                    "    attr1 = String()",
                    "\n",
                ]
            ),
        )


if __name__ == "__main__":
    testlib.unittest_main()



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Classes used to build a schema."""

from typing import Optional, List
from warnings import warn
from copy import copy

from typing import Any, Generator, Tuple, Type, Union, Dict, Sequence, Iterable, Set, cast, TypeVar

from logilab.common import attrdict, deprecation

import yams.types as yams_types

from yams import (
    BASE_TYPES,
    MARKER,
    BadSchemaDefinition,
    KNOWN_METAATTRIBUTES,
    DEFAULT_ETYPEPERMS,
    DEFAULT_RELPERMS,
    DEFAULT_ATTRPERMS,
    DEFAULT_COMPUTED_ATTRPERMS,
)
from yams.constraints import (
    SizeConstraint,
    UniqueConstraint,
    BaseConstraint,
    StaticVocabularyConstraint,
    FORMAT_CONSTRAINT,
)
from yams.schema import RelationDefinitionSchema

Defined = Dict[Union[str, Tuple[str, str, str]], Union["autopackage", "Definition"]]

__docformat__: str = "restructuredtext en"

# will be modified by the yams'reader when schema is
# beeing read
PACKAGE: str = "<builtin>"


__all__: Tuple[str, ...] = (
    "EntityType",
    "RelationType",
    "RelationDefinition",
    "SubjectRelation",
    "ObjectRelation",
    "RichString",
) + tuple(BASE_TYPES)

# EntityType properties
ETYPE_PROPERTIES: Tuple[str, ...] = ("description", "__permissions__", "__unique_together__")

# RelationType properties. Don't put description inside, handled specifically
RTYPE_PROPERTIES: Tuple[str, ...] = ("symmetric", "inlined", "fulltext_container")

# RelationDefinition properties have to be computed dynamically since new ones
# may be added at runtime


def _RELATION_DEFINITION_PROPERTIES() -> Tuple[str, ...]:
    base: Set[str] = RelationDefinitionSchema.ALL_PROPERTIES()

    # infered is an internal property and should not be specified explicitly
    base.remove("infered")

    # replace permissions by __permissions__ as it's spelled that way in schema
    # definition files
    base.remove("permissions")
    base.add("__permissions__")

    return tuple(base)


_RDEF_PROPERTIES = deprecation.renamed(
    old_name="_RDEF_PROPERTIES", new_function=_RELATION_DEFINITION_PROPERTIES
)


# regroup all rtype/rdef properties as they may be defined one on each other in
# some cases


def _RELATION_PROPERTIES() -> Tuple[str, ...]:
    return RTYPE_PROPERTIES + _RELATION_DEFINITION_PROPERTIES()


_REL_PROPERTIES = deprecation.renamed(old_name="_REL_PROPERTIES", new_function=_RELATION_PROPERTIES)


# pre 0.37 backward compat
RDEF_PROPERTIES = ()  # stuff added here is also added to underlying dict, nevermind


CREATION_RANK: int = 0


def _add_constraint(kwargs: Dict[str, Any], constraint: BaseConstraint) -> None:
    """Add constraint to param kwargs."""
    constraints: List[BaseConstraint] = kwargs.setdefault("constraints", [])

    for number, existing_constraint in enumerate(constraints):
        if existing_constraint.__class__ is constraint.__class__:
            constraints[number] = constraint

            return

    constraints.append(constraint)


@deprecation.argument_renamed(old_name="rdef", new_name="relation_definition")
@deprecation.argument_renamed(old_name="insertidx", new_name="insert_index")
def _add_relation(
    relations: List["RelationDefinition"],
    relation_definition: "RelationDefinition",
    name: Optional[yams_types.DefinitionName] = None,
    insert_index: Optional[int] = None,
) -> None:
    """Add relation (param relation_definition) to list of relations (param relations)."""
    if name is not None:
        relation_definition.name = name

    if insert_index is None:
        insert_index = len(relations)

    cast(int, insert_index)

    relations.insert(insert_index, relation_definition)

    if getattr(relation_definition, "metadata", {}):
        # mypy: "RelationDefinition" has no attribute "metadata"
        # dynamic attribute tested in the if
        for meta_name, value in relation_definition.metadata.items():  # type: ignore
            assert meta_name in KNOWN_METAATTRIBUTES

            insert_index += 1  # insert meta after main
            meta_rel_name: str = "_".join(((name or relation_definition.name), meta_name))

            _add_relation(relations, value, meta_rel_name, insert_index)


_KwargsKeyType = TypeVar("_KwargsKeyType")


def _check_kwargs(kwargs: Dict[_KwargsKeyType, Any], attributes: Sequence[_KwargsKeyType]) -> None:
    """Check that all keys of kwargs are actual attributes."""
    for key in kwargs:
        if key not in attributes:
            raise BadSchemaDefinition("no such property %r in %r" % (key, attributes))


@deprecation.argument_renamed(old_name="fromobj", new_name="from_object")
@deprecation.argument_renamed(old_name="toobj", new_name="to_object")
def _copy_attributes(from_object: Any, to_object: Any, attributes: Iterable[str]) -> None:
    for attribute in attributes:
        value = getattr(from_object, attribute, MARKER)

        if value is MARKER:
            continue

        object_value = getattr(to_object, attribute, MARKER)

        if object_value is not MARKER and value != object_value:
            relation_name = getattr(to_object, "name", None) or to_object.__name__

            raise BadSchemaDefinition(
                "conflicting values %r/%r for property %s of relation %r"
                % (object_value, value, attribute, relation_name)
            )

        setattr(to_object, attribute, value)


def register_base_types(schema: yams_types.Schema) -> None:
    """add base (final) entity types to the given schema"""
    for entity_type in BASE_TYPES:
        entity_definition = EntityType(name=entity_type)

        schema.add_entity_type(entity_definition)


# first class schema definition objects #######################################


class autopackage(type):
    def __new__(
        mcs: "Type[autopackage]", name: str, bases: Tuple, classdict: Dict[str, Any]
    ) -> Any:

        classdict["package"] = PACKAGE

        return super(autopackage, mcs).__new__(mcs, name, bases, classdict)


class Definition(object, metaclass=autopackage):
    """Abstract class for entity / relation definition classes."""

    meta = MARKER
    description = MARKER
    __permissions__ = MARKER

    def __init__(self, name=None) -> None:
        self.name: yams_types.DefinitionName = yams_types.DefinitionName(
            name or getattr(self, "name", None) or self.__class__.__name__
        )

        if self.__doc__:
            self.description = " ".join(self.__doc__.split())
            cast(str, self.description)

    def __repr__(self) -> str:
        return "<%s %r @%x>" % (self.__class__.__name__, self.name, id(self))

    @classmethod
    def expand_type_definitions(cls: Type["Definition"], defined: Defined) -> None:
        """Schema building step 1: register definition objects by adding them
        to the `defined` dictionnary.
        """
        raise NotImplementedError()

    @classmethod
    def expand_relation_definitions(
        cls: Type["Definition"], defined: Defined, schema: yams_types.Schema
    ) -> None:
        """Schema building step 2: register all relations definition,
        expanding wildcard if necessary.
        """
        raise NotImplementedError()

    def get_permissions(self, final: bool = False) -> yams_types.Permissions:
        if self.__permissions__ is MARKER:
            if final:
                return DEFAULT_ATTRPERMS

            return DEFAULT_RELPERMS

        return self.__permissions__

    @classmethod
    @deprecation.argument_renamed(old_name="perms", new_name="permissions")
    def set_permissions(cls: Type["Definition"], permissions: yams_types.Permissions) -> None:
        cls.__permissions__ = permissions


# classes used to define relationships within entity type classes ##################

# has to be defined before the metadefinition metaclass which "isinstance" this
# class
@deprecation.attribute_renamed(old_name="etype", new_name="entity_type")
class ObjectRelation:
    __permissions__ = MARKER
    cardinality = MARKER
    constraints = MARKER

    @deprecation.argument_renamed(old_name="etype", new_name="entity_type")
    def __init__(self, entity_type: str, override: bool = False, **kwargs) -> None:
        if self.__class__.__name__ == "ObjectRelation":
            warn(
                "[yams 0.29] ObjectRelation is deprecated, " "use RelationDefinition subclass",
                DeprecationWarning,
                stacklevel=2,
            )

        global CREATION_RANK
        CREATION_RANK += 1
        self.creation_rank: int = CREATION_RANK

        self.package: str = PACKAGE
        self.name: str = "<undefined>"
        self.entity_type: str = entity_type
        self.override: bool = override

        if self.constraints:
            self.constraints = list(self.constraints)

        if kwargs.pop("meta", None):
            warn("[yams 0.37.0] meta is deprecated", DeprecationWarning, stacklevel=3)

        try:
            _check_kwargs(kwargs, _RELATION_PROPERTIES())
        except BadSchemaDefinition as bad:
            # XXX (auc) bad field name + required attribute can lead there
            # instead of schema.py ~ 920
            bad_schema_definition = BadSchemaDefinition(
                "%s in relation to entity %r (also is %r defined ? "
                "(check two lines above in the backtrace))" % (bad.args, entity_type, entity_type)
            )
            # mypy: "BadSchemaDefinition" has no attribute "tb_offset"
            # hack to transport information
            bad_schema_definition.tb_offset = 2  # type: ignore

            raise bad_schema_definition

        self.__dict__.update(kwargs)

    def __repr__(self) -> str:
        return "%(name)s %(entity_type)s" % self.__dict__


class SubjectRelation(ObjectRelation):
    uid = MARKER
    indexed = MARKER
    fulltextindexed = MARKER
    internationalizable = MARKER
    default = MARKER

    def __repr__(self) -> str:
        return "%(entity_type)s %(name)s" % self.__dict__


class AbstractTypedAttribute(SubjectRelation):
    """AbstractTypedAttribute is not directly instantiable

    subclasses must provide a <entity_type> attribute to be instantiable
    """

    def __init__(
        self,
        metadata: Optional[Dict[str, "AbstractTypedAttribute"]] = None,
        required: bool = False,
        maxsize: Optional[int] = None,
        formula=MARKER,
        vocabulary: Optional[List[str]] = None,
        unique: Optional[bool] = None,
        override: bool = False,
        **kwargs,
    ):

        # Store metadata
        if metadata is None:
            metadata = {}
        self.metadata: Dict[Any, "AbstractTypedAttribute"] = metadata

        # transform "required" into "cardinality"
        if required:
            cardinality = "11"
        else:
            cardinality = "?1"
        kwargs["cardinality"] = cardinality

        # transform maxsize into SizeConstraint
        if maxsize is not None:
            _add_constraint(kwargs, SizeConstraint(max=maxsize))

        # formula
        self.formula = formula

        # transform vocabulary into StaticVocabularyConstraint
        if vocabulary is not None:
            self.set_vocabulary(vocabulary, kwargs)

        # transform unique into UniqueConstraint
        if unique:
            _add_constraint(kwargs, UniqueConstraint())

        # use the entity_type attribute provided by subclasses
        kwargs["override"] = override
        super(AbstractTypedAttribute, self).__init__(self.entity_type, **kwargs)

        # reassign creation rank
        #
        # Main attribute are marked as created before it's metadata.
        # order in meta data is preserved.
        if self.metadata:
            meta: List[AbstractTypedAttribute] = sorted(
                metadata.values(), key=lambda x: x.creation_rank
            )

            if meta[0].creation_rank < self.creation_rank:
                _previous: AbstractTypedAttribute = self

                for _next in meta:
                    if _previous.creation_rank < _next.creation_rank:
                        break

                    _previous.creation_rank, _next.creation_rank = (
                        _next.creation_rank,
                        _previous.creation_rank,
                    )
                    _next = _previous

    def set_vocabulary(self, vocabulary: List[str], kwargs=None) -> None:
        if kwargs is None:
            kwargs = self.__dict__

        # constraints = kwargs.setdefault('constraints', [])
        _add_constraint(kwargs, StaticVocabularyConstraint(vocabulary))

        if self.__class__.__name__ == "String":  # XXX
            max_size = max(len(x) for x in vocabulary)

            _add_constraint(kwargs, SizeConstraint(max=max_size))

    def __repr__(self) -> str:
        return "<%(name)s(%(entity_type)s)>" % self.__dict__


@deprecation.argument_renamed(old_name="etype", new_name="entity_type")
def make_type(entity_type: str) -> Type[AbstractTypedAttribute]:
    """create a python class for a Yams base type.

    Notice it is now possible to create a specific type with user-defined
    behaviour, e.g.:

        Geometry = make_type('Geometry') # (c.f. postgis)

    will allow the use of:

        Geometry(geom_type='POINT')

    in a Yams schema, provided in this example that `geom_type` is specified to
    the :func:`yams.register_base_type` function which should be called prior to
    make_type.
    """
    assert entity_type in BASE_TYPES
    return type(entity_type, (AbstractTypedAttribute,), {"entity_type": entity_type})


# build a specific class for each base type
class String(AbstractTypedAttribute):
    entity_type: str = "String"


class Password(AbstractTypedAttribute):
    entity_type: str = "Password"


class Bytes(AbstractTypedAttribute):
    entity_type: str = "Bytes"


class Int(AbstractTypedAttribute):
    entity_type: str = "Int"


class BigInt(AbstractTypedAttribute):
    entity_type: str = "BigInt"


class Float(AbstractTypedAttribute):
    entity_type: str = "Float"


class Boolean(AbstractTypedAttribute):
    entity_type: str = "Boolean"


class Decimal(AbstractTypedAttribute):
    entity_type: str = "Decimal"


class Time(AbstractTypedAttribute):
    entity_type: str = "Time"


class Date(AbstractTypedAttribute):
    entity_type: str = "Date"


class Datetime(AbstractTypedAttribute):
    entity_type: str = "Datetime"


class TZTime(AbstractTypedAttribute):
    entity_type: str = "TZTime"


class TZDatetime(AbstractTypedAttribute):
    entity_type: str = "TZDatetime"


class Interval(AbstractTypedAttribute):
    entity_type: str = "Interval"


# provides a RichString factory for convenience
def RichString(
    default_format: str = "text/plain",
    format_constraints: Optional[List[BaseConstraint]] = None,
    required: bool = False,
    maxsize: Optional[int] = None,
    formula=MARKER,
    vocabulary: Optional[List[str]] = None,
    unique: Optional[bool] = None,
    override: bool = False,
    **kwargs,
):
    """RichString is a convenience attribute type for attribute containing text
    in a format that should be specified in another attribute.

    The following declaration::

      class Card(EntityType):
          content = RichString(fulltextindexed=True, default_format='text/rest')

    is equivalent to::

      class Card(EntityType):
          content_format = String(internationalizable=True,
                                  default='text/rest', constraints=[FORMAT_CONSTRAINT])
          content  = String(fulltextindexed=True)
    """
    format_args = {"default": default_format, "maxsize": 50}

    if format_constraints is None:
        format_args["constraints"] = [FORMAT_CONSTRAINT]
    else:
        format_args["constraints"] = format_constraints

    # mypy: Argument 2 to "String" has incompatible type "**Dict[str, object]"; expected
    # mypy: "Optional[bool]"
    # really looks like mypy is failing on AbstractTypedAttribute constructor here
    meta: Dict[str, AbstractTypedAttribute] = {
        "format": String(internationalizable=True, **format_args)  # type: ignore
    }

    return String(
        metadata=meta,
        required=required,
        maxsize=maxsize,
        formula=formula,
        vocabulary=vocabulary,
        unique=unique,
        override=override,
        **kwargs,
    )


# other schema definition classes ##############################################


class metadefinition(autopackage):
    """Metaclass that builds the __relations__ attribute of EntityType's
    subclasses.
    """

    stacklevel = 3

    def __new__(
        mcs: "Type[metadefinition]", name: str, bases: Tuple, classdict: Dict[str, Any]
    ) -> Any:

        # Move (any) relation from the class dict to __relations__ attribute
        relations_list: List = classdict.setdefault("__relations__", [])
        relations: Dict[str, Any] = dict(
            (relation_definition.name, relation_definition)
            for relation_definition in relations_list
        )

        for relation_name, relation_definition in list(classdict.items()):
            if isinstance(relation_definition, ObjectRelation):
                # relation's name **must** be removed from class namespace
                # to avoid conflicts with instance's potential attributes
                del classdict[relation_name]

                relations[relation_name] = relation_definition

        # handle logical inheritance
        if "__specializes_schema__" in classdict:
            specialized = bases[0]
            classdict["__specializes__"] = specialized.__name__

            if "__specialized_by__" not in specialized.__dict__:
                specialized.__specialized_by__ = []

            specialized.__specialized_by__.append(name)

        # Initialize processed class
        class_definition = super(metadefinition, mcs).__new__(mcs, name, bases, classdict)

        for relation_name, relation_definition in relations.items():
            _add_relation(class_definition.__relations__, relation_definition, relation_name)

        # take base classes'relations into account
        for base in bases:
            for relation_definition in getattr(base, "__relations__", ()):
                if (
                    relation_definition.name not in relations
                    or not relations[relation_definition.name].override
                ):
                    if isinstance(relation_definition, RelationDefinition):
                        relation_definition = copy(relation_definition)

                        if relation_definition.subject == base.__name__:
                            relation_definition.subject = name

                        if relation_definition.object == base.__name__:
                            relation_definition.object = name

                    relations_list.append(relation_definition)
                else:
                    relations[
                        relation_definition.name
                    ].creation_rank = relation_definition.creation_rank

        # sort relations by creation rank
        class_definition.__relations__ = sorted(relations_list, key=lambda r: r.creation_rank)

        return class_definition


class EntityType(Definition, metaclass=metadefinition):
    # :FIXME reader magic forbids to define a docstring...
    #  an entity has attributes and can be linked to other entities by
    #  relations. Both entity attributes and relationships are defined by
    #  class attributes.
    #
    #  kwargs keys must have values in ETYPE_PROPERTIES
    #
    #  Example:
    #
    #  >>> class Project(EntityType):
    #  ...     name = String()
    #  >>>
    #
    #  After instanciation, EntityType can we altered with dedicated class methods:
    #
    #  .. currentmodule:: yams.buildobjs
    #
    #   .. automethod:: EntityType.extend
    #   .. automethod:: EntityType.add_relation
    #   .. automethod:: EntityType.insert_relation_after
    #   .. automethod:: EntityType.remove_relation
    #   .. automethod:: EntityType.get_relation
    #   .. automethod:: EntityType.get_relations

    __permissions__: yams_types.Permissions = DEFAULT_ETYPEPERMS

    def __init__(self, name: Optional[str] = None, **kwargs) -> None:
        super(EntityType, self).__init__(name)

        _check_kwargs(kwargs, ETYPE_PROPERTIES)

        self.__dict__.update(kwargs)
        self.specialized_type: Optional[str] = self.__class__.__dict__.get("__specializes__")

    def __str__(self) -> str:
        return "entity type %r" % self.name

    @property
    def specialized_by(self) -> List[str]:
        return self.__class__.__dict__.get("__specialized_by__", [])

    @classmethod
    def expand_type_definitions(cls: Type["EntityType"], defined: Defined) -> None:
        """Schema building step 1: register definition objects by adding
        them to the `defined` dictionnary.
        """
        name: str = getattr(cls, "name", cls.__name__)

        assert cls is not defined.get(name), "duplicate registration: %s" % name
        assert (
            name not in defined
        ), "type '%s' was already defined here %s, new definition here %s" % (
            name,
            defined[name].__module__,
            cls,
        )

        # mypy: "Type[EntityType]" has no attribute "_defined"
        # dynamic attribute
        # XXX may be used later (eg .add_relation())
        cls._defined: Defined = defined  # type: ignore

        defined[name] = cls

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        for relation in cls.__relations__:  # type: ignore
            cls._ensure_relation_type(relation)

    @classmethod
    def _ensure_relation_type(cls: Type["EntityType"], relation: ObjectRelation) -> bool:
        """Check the type the relation

        return False if the class is not yet finalized
        (XXX raise excep instead ?)"""

        relation_type = RelationType(relation.name)

        _copy_attributes(relation, relation_type, RTYPE_PROPERTIES)

        # assert hasattr(cls, '_defined'), "Type definition for %s not yet expanded.
        # you can't register new type through it" % cls

        if not hasattr(cls, "_defined"):
            return False

        # mypy: "Type[EntityType]" has no attribute "_defined"
        # dynamically set attribute
        defined = cls._defined  # type: ignore

        if relation.name in defined:
            _copy_attributes(relation_type, defined[relation.name], RTYPE_PROPERTIES)
        else:
            defined[relation.name] = relation_type

        return True

    @classmethod
    def expand_relation_definitions(
        cls: Type["EntityType"], defined: Defined, schema: yams_types.Schema
    ) -> None:
        """schema building step 2:

        register all relations definition, expanding wildcards if necessary
        """
        order: int = 1
        name: str = getattr(cls, "name", cls.__name__)
        relation_definitions_properties: Tuple[str, ...] = _RELATION_DEFINITION_PROPERTIES()

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        for relation in cls.__relations__:  # type: ignore
            if isinstance(relation, SubjectRelation):
                relation_definition = RelationDefinition(
                    subject=name,
                    name=relation.name,
                    object=relation.entity_type,
                    order=order,
                    package=relation.package,
                )

                _copy_attributes(relation, relation_definition, relation_definitions_properties)

            elif isinstance(relation, ObjectRelation):
                relation_definition = RelationDefinition(
                    subject=relation.entity_type,
                    name=relation.name,
                    object=name,
                    order=order,
                    package=relation.package,
                )

                _copy_attributes(relation, relation_definition, relation_definitions_properties)

            elif isinstance(relation, RelationDefinition):
                relation_definition = relation
            else:
                raise BadSchemaDefinition("dunno how to handle %s" % relation)

            order += 1
            relation_definition._add_relations(defined, schema)

    # methods that can be used to extend an existant schema definition ########

    @classmethod
    @deprecation.attribute_renamed(
        old_name="othermetadefcls", new_name="other_meta_definition_class"
    )
    def extend(cls: Type["EntityType"], other_meta_definition_class) -> None:
        """add all relations of ``other_meta_definition_class`` to the current class"""
        for relation_definition in other_meta_definition_class.__relations__:
            cls.add_relation(relation_definition)

    @classmethod
    @deprecation.argument_renamed(old_name="rdef", new_name="relation_definition")
    def add_relation(
        cls: Type["EntityType"], relation_definition: ObjectRelation, name: Optional[str] = None
    ) -> None:
        """Add ``relation_definition`` relation to the class"""
        if name:
            relation_definition.name = name

        if cls._ensure_relation_type(relation_definition):
            # mypy: "Type[EntityType]" has no attribute "__relations__"
            # dynamically set attribute, full yams magic
            _add_relation(cls.__relations__, relation_definition, name)  # type: ignore

            # mypy: "Type[EntityType]" has no attribute "_defined"
            # dynamically set attribute
            if (
                getattr(relation_definition, "metadata", {})
                and relation_definition not in cls._defined  # type: ignore
            ):
                for meta_name in relation_definition.metadata:  # type: ignore
                    format_attr_name = "_".join(((name or relation_definition.name), meta_name))
                    relation_definition = next(cls.get_relations(format_attr_name))
                    cls._ensure_relation_type(relation_definition)

        else:
            # mypy: "Type[EntityType]" has no attribute "__relations__"
            # dynamically set attribute, full yams magic
            _add_relation(cls.__relations__, relation_definition, name=name)  # type: ignore

    @classmethod
    @deprecation.argument_renamed(old_name="rdef", new_name="relation_definition")
    @deprecation.argument_renamed(old_name="afterrelname", new_name="after_relation_name")
    def insert_relation_after(
        cls: Type["EntityType"],
        after_relation_name: str,
        name: str,
        relation_definition: ObjectRelation,
    ) -> None:
        """Add ``relation_definition`` relation to the class right after another"""
        # FIXME change order of arguments to relation_definition, name, after_relation_name ?
        relation_definition.name = name
        cls._ensure_relation_type(relation_definition)

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        number = 0
        for number, rel in enumerate(cls.__relations__):  # type: ignore
            if rel.name == after_relation_name:
                break

        else:
            raise BadSchemaDefinition("can't find %s relation on %s" % (after_relation_name, cls))

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        _add_relation(cls.__relations__, relation_definition, name, number + 1)  # type: ignore

    @classmethod
    def remove_relation(cls: Type["EntityType"], name: str) -> None:
        """Remove relation from the class"""

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        for relation_definition in cls.get_relations(name):
            cls.__relations__.remove(relation_definition)  # type: ignore

    @classmethod
    def get_relations(cls: Type["EntityType"], name: str) -> Generator[ObjectRelation, Any, None]:
        """Iterate over relations definitions that match the ``name`` parameters

        It may iterate multiple definitions when the class is both object and
        sujet of a relation:
        """

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        for relation_definition in cls.__relations__[:]:  # type: ignore
            if relation_definition.name == name:
                yield relation_definition

    @classmethod
    def get_relation(cls: Type["EntityType"], name: str) -> ObjectRelation:
        """Return relation definitions by name. Fails if there is multiple one."""
        relations: Tuple[ObjectRelation, ...] = tuple(cls.get_relations(name))

        assert len(relations) == 1, "can't use get_relation for relation with multiple definitions"

        return relations[0]


class RelationType(Definition):
    symmetric = MARKER
    inlined = MARKER
    fulltext_container = MARKER
    rule = MARKER

    def __init__(self, name: Optional[str] = None, **kwargs) -> None:
        """kwargs must have values in RTYPE_PROPERTIES"""
        super(RelationType, self).__init__(name)

        if kwargs.pop("meta", None):
            warn("[yams 0.37] meta is deprecated", DeprecationWarning, stacklevel=2)

        _check_kwargs(kwargs, RTYPE_PROPERTIES + ("description", "__permissions__"))

        self.__dict__.update(kwargs)

    def __str__(self) -> str:
        return "relation type %r" % self.name

    @classmethod
    def expand_type_definitions(cls: Type["RelationType"], defined: Defined) -> None:
        """schema building step 1:

        register definition objects by adding them to the `defined` dictionnary
        """
        name: str = getattr(cls, "name", cls.__name__)

        if cls.__doc__ and not cls.description:
            cls.description = " ".join(cls.__doc__.split())

        if name in defined:
            if defined[name].__class__ is not RelationType:
                raise BadSchemaDefinition("duplicated relation type for %s" % name)

            # relation type created from a relation definition, override it
            all_properties = _RELATION_PROPERTIES() + ("subject", "object")

            _copy_attributes(defined[name], cls, all_properties)

        defined[name] = cls

    @classmethod
    def expand_relation_definitions(
        cls: Type["RelationType"], defined: Defined, schema: yams_types.Schema
    ) -> None:
        """schema building step 2:

        register all relations definition, expanding wildcard if necessary
        """
        name: str = getattr(cls, "name", cls.__name__)

        if getattr(cls, "subject", None) and getattr(cls, "object", None):
            # mypy: "Type[RelationType]" has no attribute "subject"
            # mypy: "Type[RelationType]" has no attribute "object"
            # dynamically set attributes
            relation_definition = RelationDefinition(
                subject=cls.subject,  # type: ignore
                name=name,  # type: ignore
                object=cls.object,  # type: ignore
            )

            relation_definition._add_relations(defined, schema)
            _copy_attributes(cls, relation_definition, _RELATION_DEFINITION_PROPERTIES())


class ComputedRelation(RelationType):
    __permissions__ = MARKER

    def __init__(self, name: Optional[str] = None, rule=None, **kwargs) -> None:
        if rule is not None:
            self.rule = rule

        super(ComputedRelation, self).__init__(name, **kwargs)


class RelationDefinition(Definition):
    # FIXME reader magic forbids to define a docstring...
    # """a relation is defined by a name, the entity types that can be
    # subject or object the relation, the cardinality, the constraints
    # and the symmetric property.
    # """

    subject = MARKER
    object = MARKER
    cardinality = MARKER
    constraints = MARKER
    symmetric = MARKER
    inlined = MARKER
    formula = MARKER

    def __init__(
        self,
        subject: Optional[str] = None,
        name: Optional[str] = None,
        object: Optional[str] = None,
        package: Optional[str] = None,
        **kwargs,
    ) -> None:
        """kwargs keys must have values in _RELATION_DEFINITION_PROPERTIES()"""
        if subject:
            self.subject = subject
        else:
            self.subject = self.__class__.subject

        if object:
            self.object = object
        else:
            self.object = self.__class__.object

        super(RelationDefinition, self).__init__(name)

        global CREATION_RANK
        CREATION_RANK += 1
        self.creation_rank: int = CREATION_RANK

        self.package: str
        if package is not None:
            self.package = package
        elif self.package == "<builtin>":
            self.package = PACKAGE

        if kwargs.pop("meta", None):
            warn("[yams 0.37] meta is deprecated", DeprecationWarning)

        relation_definitions_properties: Tuple[str, ...] = _RELATION_DEFINITION_PROPERTIES()

        _check_kwargs(kwargs, relation_definitions_properties)
        _copy_attributes(attrdict(**kwargs), self, relation_definitions_properties)

        if self.constraints:
            self.constraints = list(self.constraints)

    def __str__(self) -> str:
        return "relation definition (%(subject)s %(name)s %(object)s)" % self.__dict__

    @classmethod
    def expand_type_definitions(cls: Type["RelationDefinition"], defined: Defined) -> None:
        """schema building step 1:

        register definition objects by adding them to the `defined` dictionnary
        """
        name: str = getattr(cls, "name", cls.__name__)
        relation_type: RelationType = RelationType(name)

        _copy_attributes(cls, relation_type, RTYPE_PROPERTIES)

        if name in defined:
            _copy_attributes(relation_type, defined[name], RTYPE_PROPERTIES)
        else:
            defined[name] = relation_type

        # subject and object in defined's keys are only strings not tuples
        if isinstance(cls.subject, tuple):
            subjects = cls.subject
        else:
            subjects = (cls.subject,)

        if isinstance(cls.object, tuple):
            objects = cls.object
        else:
            objects = (cls.object,)

        for subject in subjects:
            for object in objects:
                key = (subject, name, object)
                if key in defined:
                    raise BadSchemaDefinition(
                        "duplicated relation definition (%s) %s (%s.%s)"
                        % (defined[key], key, cls.__module__, cls.__name__)
                    )
                defined[key] = cls

        # XXX keep this for bw compat
        defined[(cls.subject, name, cls.object)] = cls

    @classmethod
    def expand_relation_definitions(
        cls: Type["RelationDefinition"], defined: Defined, schema: yams_types.Schema
    ) -> None:
        """schema building step 2:

        register all relations definition, expanding wildcard if necessary
        """
        assert cls.subject and cls.object, "%s; check the schema (%s, %s)" % (
            cls,
            cls.subject,
            cls.object,
        )
        cls()._add_relations(defined, schema)

    def _add_relations(self, defined: Defined, schema: yams_types.Schema) -> None:
        name: str = getattr(self, "name", self.__class__.__name__)
        relation_type: Union[autopackage, Definition] = defined[name]
        relation_definitions_properties: Tuple[str, ...] = _RELATION_DEFINITION_PROPERTIES()

        # copy relation definition attributes set on the relation type, beside
        # description
        _copy_attributes(
            relation_type, self, set(relation_definitions_properties) - set(("description",))
        )

        # process default cardinality and constraints if not set yet
        cardinality = self.cardinality
        if cardinality is MARKER:
            if self.object in BASE_TYPES:
                self.cardinality = "?1"
            else:
                self.cardinality = "**"
        else:
            assert len(cardinality) == 2
            assert cardinality[0] in "1?+*"
            assert cardinality[1] in "1?+*"

        if not self.constraints:
            self.constraints = ()

        relation_schema = schema.relation_schema_for(yams_types.DefinitionName(name))
        if relation_schema.rule:
            raise BadSchemaDefinition(
                'Cannot add relation definition "{}" because an '
                "homonymous computed relation already exists "
                'with rule "{}"'.format(relation_schema.type, relation_schema.rule)
            )

        if self.__permissions__ is MARKER:
            final: bool = next(iter(_actual_types(schema, self.object))) in BASE_TYPES

            if final:
                if self.formula is not MARKER:
                    permissions = DEFAULT_COMPUTED_ATTRPERMS
                else:
                    permissions = DEFAULT_ATTRPERMS
            else:
                permissions = DEFAULT_RELPERMS
        else:
            permissions = self.__permissions__

        for subject in _actual_types(schema, self.subject):
            for object in _actual_types(schema, self.object):
                relation_definition = RelationDefinition(
                    subject, name, object, __permissions__=permissions, package=self.package
                )

                _copy_attributes(self, relation_definition, relation_definitions_properties)
                schema.add_relation_def(relation_definition)


@deprecation.argument_renamed(old_name="etype", new_name="entity_type")
def _actual_types(
    schema: yams_types.Schema, entity_type: Union[str, list, Tuple[Any, Any]]
) -> Union[Generator[Any, Any, None], Tuple[Any], Tuple, List, Any]:
    if entity_type == "*":
        yield from (eschema.type for eschema in schema.entities() if not eschema.final)
    elif isinstance(entity_type, (list, tuple)):
        yield from entity_type
    elif isinstance(entity_type, str):
        yield entity_type
    else:
        raise RuntimeError(
            "Entity types must be strings or list/tuples of strings. "
            'SubjectRelation(Foo) is wrong, SubjectRelation("Foo") is correct. '
            "Hence, %r is not acceptable." % entity_type
        )



"""some yams command line tools"""

import sys
from os.path import exists, join, dirname
from tempfile import NamedTemporaryFile
from subprocess import Popen
from traceback import extract_tb

from logilab.common.configuration import Configuration

from yams import __version__ as version
from yams import schema2dot
from yams.reader import SchemaLoader
from typing import Optional, Union


def _error(
    file: Optional[str] = None, line: Optional[Union[int, str]] = None, msg: str = ""
) -> None:
    if file is None:
        file = sys.argv[1]
    if line is None:
        line = ""
    else:
        line = str(line)
    print(":".join(("E", file, line, msg)), file=sys.stderr)


def check_schema() -> int:
    config = Configuration(
        usage="yams-check [[[...] deps] deps] apps",
        doc="Check the schema of an application.",
        version=version,
    )

    dirnames = config.load_command_line_configuration()

    if not dirnames:
        print(config.help(), file=sys.stderr)

        return 2

    for dir_ in dirnames:
        assert exists(dir_), dir_

    try:
        SchemaLoader().load(dirnames)

        return 0
    except Exception as ex:
        tb_offset = getattr(ex, "tb_offset", 0)
        _, _, traceback = sys.exc_info()
        filename, lineno, _, _ = extract_tb(traceback)[-1 - tb_offset]

        if hasattr(ex, "schema_files"):
            # mypy: "Exception" has no attribute "schema_files"
            # but we've added it above in the exception chain and we test it before
            filename = ", ".join(ex.schema_files)  # type: ignore

        _error(filename, lineno, "%s -> %s" % (ex.__class__.__name__, ex))

        return 2


def schema_image() -> int:
    options = [
        (
            "output-file",
            {
                "type": "file",
                "default": None,
                "metavar": "<file>",
                "short": "o",
                "help": "output image file",
            },
        ),
        (
            "viewer",
            {
                "type": "string",
                "default": "rsvg-view",
                "short": "w",
                "metavar": "<cmd>",
                "help": "command use to view the generated file (empty for none)",
            },
        ),
        (
            "lib-dir",
            {
                "type": "string",
                "short": "L",
                "metavar": "<dir>",
                "help": "directory to look for schema dependancies",
            },
        ),
    ]

    config = Configuration(
        options=options,
        usage="yams-view [-L <lib_dir> | [[[...] deps] deps]] apps",
        version=version,
    )

    dirnames = config.load_command_line_configuration()
    lib_dir = config["lib-dir"]

    assert lib_dir is not None

    if lib_dir is not None:
        app_dir = dirnames[-1]

        from cubicweb.cwconfig import CubicWebConfiguration as cwcfg

        packages = cwcfg.expand_cubes(dirnames)
        packages = cwcfg.reorder_cubes(packages)
        packages = [pkg for pkg in packages if pkg != app_dir]
    elif False:
        glob = globals().copy()

        exec(open(join(app_dir, "__pkginfo__.py")).read(), glob)

        # dirnames = [ join(lib_dir,dep) for dep in glob['__use__']]+dirnames
        packages = [dep for dep in glob["__use__"]]

    for dir_ in dirnames:
        assert exists(dir_), dir_

    import cubicweb

    cubicweb_dir = dirname(cubicweb.__file__)

    schm_ldr = SchemaLoader()

    class MockConfig:
        def packages(self):
            return packages

        def packages_path(self):
            return packages

        def schemas_lib_dir(self):
            return join(cubicweb_dir, "schemas")

        # def apphome(self):
        #    return lib_dir
        apphome = dirnames[0]

        def appid(self):
            "bob"

    print(MockConfig().packages())

    # mypy: Argument 1 to "load" of "SchemaLoader" has incompatible type "MockConfig";
    # mypy: expected "Sequence[Tuple[Any, str]]"
    # mocking
    schema = schm_ldr.load(MockConfig())  # type: ignore

    out, viewer = config["output-file"], config["viewer"]

    if out is None:
        tmp_file = NamedTemporaryFile(suffix=".svg")
        out = tmp_file.name

    schema2dot.schema2dot(
        schema,
        out,  # size=size,
        skiptypes=("Person", "AbstractPerson", "Card", "AbstractCompany", "Company", "Division"),
    )

    if viewer:
        p = Popen((viewer, out))
        p.wait()

    return 0



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with yams. If not, see <https://www.gnu.org/licenses/>.
"""specialization tests
"""
from logilab.common.testlib import TestCase, unittest_main

from yams.reader import build_schema_from_namespace
from yams.buildobjs import EntityType, String, SubjectRelation, RelationDefinition


def build_schema():
    class Person(EntityType):
        firstname = String()
        knows = SubjectRelation("Person")
        works_for = SubjectRelation("Company")

    class Student(Person):
        __specializes_schema__ = True

    class Company(EntityType):
        name = String()

    class SubCompany(Company):
        __specializes_schema__ = True

    class Division(Company):
        __specializes_schema__ = True
        division_of = SubjectRelation("Company")

    class SubDivision(Division):
        __specializes_schema__ = True

    # This class doesn't extend the schema
    class SubSubDivision(SubDivision):
        pass

    class custom_attr(RelationDefinition):
        subject = "Person"
        object = "String"
        __permissions__ = {"read": ("managers",), "add": ("managers",), "update": ("managers",)}

    return build_schema_from_namespace(locals().items())


class SpecializationTC(TestCase):
    def setUp(self):
        self.schema = build_schema()

    def test_schema_specialization(self):
        schema = self.schema
        # company
        company = schema.entity_schema_for("Company")
        self.assertEqual(company.specializes(), None)
        # division
        division = schema.entity_schema_for("Division")
        self.assertEqual(division.specializes().type, "Company")
        # subdivision
        subdivision = schema.entity_schema_for("SubDivision")
        self.assertEqual(subdivision.specializes().type, "Division")
        # subsubdivision
        subsubdivision = schema.entity_schema_for("SubSubDivision")
        self.assertEqual(subsubdivision.specializes(), None)

    def test_ancestors(self):
        schema = self.schema
        # company
        company = schema.entity_schema_for("Company")
        self.assertEqual(company.ancestors(), [])
        # division
        division = schema.entity_schema_for("Division")
        self.assertEqual(division.ancestors(), ["Company"])
        # subdivision
        subdivision = schema.entity_schema_for("SubDivision")
        self.assertEqual(subdivision.ancestors(), ["Division", "Company"])
        # subsubdivision
        subsubdivision = schema.entity_schema_for("SubSubDivision")
        self.assertEqual(subsubdivision.ancestors(), [])

    def test_specialized_by(self):
        schema = self.schema
        # company
        company = schema.entity_schema_for("Company")
        self.assertEqual(sorted(company.specialized_by(False)), ["Division", "SubCompany"])
        self.assertEqual(
            sorted(company.specialized_by(True)), ["Division", "SubCompany", "SubDivision"]
        )

        # division
        division = schema.entity_schema_for("Division")
        self.assertEqual(sorted(division.specialized_by(False)), ["SubDivision"])
        self.assertEqual(sorted(division.specialized_by(True)), ["SubDivision"])
        # subdivision
        subdivision = schema.entity_schema_for("SubDivision")
        self.assertEqual(sorted(subdivision.specialized_by(False)), [])
        # subsubdivision
        subsubdivision = schema.entity_schema_for("SubSubDivision")
        self.assertEqual(subsubdivision.specialized_by(False), [])

    def test_relations_infered(self):
        entities = [str(e) for e in self.schema.entities() if not e.final]
        relations = sorted([r for r in self.schema.relations() if not r.final])
        self.assertListEqual(
            sorted(entities),
            [
                "Company",
                "Division",
                "Person",
                "Student",
                "SubCompany",
                "SubDivision",
                "SubSubDivision",
            ],
        )
        self.assertListEqual(relations, ["division_of", "knows", "works_for"])
        expected = {
            ("Person", "Person"): False,
            ("Person", "Student"): True,
            # as Student extends Person,
            # it already has the `knows` relation
            ("Student", "Person"): False,
            ("Student", "Student"): True,
        }
        done = set()
        drschema, krschema, wrschema = relations
        for subjobj in krschema.relation_definitions:
            subject, object = subjobj
            done.add(subjobj)
            self.assertIn(subjobj, expected)
            self.assertEqual(
                krschema.relation_definition(subject, object).infered, expected[subjobj]
            )
        self.assertEqual(len(set(expected) - done), 0, "missing %s" % (set(expected) - done))
        expected = {
            ("Person", "Company"): False,
            ("Person", "Division"): True,
            ("Person", "SubDivision"): True,
            ("Person", "SubCompany"): True,
            ("Student", "Company"): False,
            ("Student", "Division"): True,
            ("Student", "SubDivision"): True,
            ("Student", "SubCompany"): True,
        }
        done = set()
        for subjobj in wrschema.relation_definitions:
            subject, object = subjobj
            done.add(subjobj)
            self.assertIn(subjobj, expected)
            self.assertEqual(
                wrschema.relation_definition(subject, object).infered, expected[subjobj]
            )
        self.assertEqual(len(set(expected) - done), 0, "missing %s" % (set(expected) - done))

        self.assertIn("custom_attr", self.schema["Student"]._subject_relations)
        self.assertEqual(
            self.schema["custom_attr"].relation_definitions[("Student", "String")].permissions,
            {"read": ("managers",), "add": ("managers",), "update": ("managers",)},
        )

    def test_remove_infered_relations(self):
        self.schema.remove_infered_definitions()
        relations = sorted([r for r in self.schema.relations() if not r.final])
        self.assertListEqual(relations, ["division_of", "knows", "works_for"])
        expected = {
            ("Person", "Person"): False,
            # as Student extends Person, it already has the `knows` relation
            ("Student", "Person"): False,
        }
        done = set()
        drschema, krschema, wrschema = relations
        for subjobj in krschema.relation_definitions:
            subject, object = subjobj
            done.add(subjobj)
            self.assertIn(subjobj, expected)
            self.assertEqual(
                krschema.relation_definition(subject, object).infered, expected[subjobj]
            )
        self.assertEqual(len(set(expected) - done), 0, "missing %s" % (set(expected) - done))
        expected = {
            ("Person", "Company"): False,
            ("Student", "Company"): False,
        }
        done = set()
        for subjobj in wrschema.relation_definitions:
            subject, object = subjobj
            done.add(subjobj)
            self.assertTrue(subjobj in expected)
            self.assertEqual(
                wrschema.relation_definition(subject, object).infered, expected[subjobj]
            )
        self.assertEqual(len(set(expected) - done), 0, "missing %s" % (set(expected) - done))

    def test_no_more_infered_relations(self):
        relation_definition = self.schema["division_of"].relation_definitions[
            "SubSubDivision", "SubCompany"
        ]
        self.assertEqual("**", relation_definition.cardinality)
        relation_definition = RelationDefinition(
            "SubSubDivision", "division_of", "SubCompany", cardinality="1*"
        )
        # ensure add_relation_def doesn't raise an error
        self.schema.add_relation_def(relation_definition)
        relation_definition = self.schema["division_of"].relation_definitions[
            "SubSubDivision", "SubCompany"
        ]
        self.assertEqual("1*", relation_definition.cardinality)
        relation_definition = self.schema["SubSubDivision"].relation_definition(
            "division_of", "subject", "SubCompany"
        )
        self.assertEqual("1*", relation_definition.cardinality)
        relation_definition = self.schema["SubCompany"].relation_definition(
            "division_of", "object", "SubSubDivision"
        )
        self.assertEqual("1*", relation_definition.cardinality)

    def test_remove_infered_relations_dont_remove_relation_type(self):
        # for the sake of this test, mark all relation_definition as infered even if that makes
        # no sense in the inheritance case (at least one won't be infered in
        # real life). However this will be the case with computed relations,
        # though we've only partial implementation in yams that can't be easily
        # tested, and they will rely on the tested behaviour of
        # remove_infered_definitions
        for relation_definition in self.schema["works_for"].relation_definitions.values():
            relation_definition.infered = True
        self.schema.remove_infered_definitions()
        self.assertIn("works_for", self.schema)


if __name__ == "__main__":
    unittest_main()



# copyright 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with yams. If not, see <https://www.gnu.org/licenses/>.
"""XML vocabularies <-> Yams schema mapping"""

from warnings import warn
from typing import Any, List, Union, Tuple, Dict


class UnsupportedVocabulary(Exception):
    pass


XYSet = Tuple[str, Union[str, Tuple[str, str]], str]


class XYRegistry:
    """registry of equivalence to translate yams into various xml schema such
    as dublin core, foaf, doap...

    """

    def __init__(self):
        self.prefixes: Dict[str, str] = {}
        self._y2x: Dict[XYSet, List[XYSet]] = {}
        self._x2y: Dict[XYSet, List[XYSet]] = {}

    def register_prefix(self, prefix: str, xmlns: str, overwrite: bool = False) -> None:
        if ":" in prefix:
            warn(
                "[yams 0.30] register_prefix arguments has been switched to " "(prefix, xmlns)",
                DeprecationWarning,
                stacklevel=2,
            )

            prefix, xmlns = xmlns, prefix

        if not overwrite and prefix in self.prefixes and self.prefixes.get(prefix) != xmlns:
            raise Exception("prefix %r already defined with different value" % prefix)

        self.prefixes[prefix] = xmlns

    def _norm_yams_key(self, yamssnippet: str) -> Tuple[str, str, str]:
        parts = yamssnippet.split(" ")

        if len(parts) == 1:
            if parts[0][0].islower():
                return "*", parts[0], "*"

            # mypy: Incompatible return value type (got "str", expected "Tuple[str, str, str]")
            # XXX or is it a bug? You'll get a str here which breaks add_equivalence
            return parts[0]  # type: ignore

        if len(parts) == 2:
            parts.append("*")

        assert len(parts) == 3

        # mypy: Incompatible return value type (got "Tuple[str, ...]",
        # mypy: expected "Tuple[str, str, str]")
        # but we know that "parts" is of len 3 because of the assert
        return tuple(parts)  # type: ignore

    def _norm_xml_key(self, xmlsnippet: str, isentity: bool = False) -> XYSet:

        parts: List[Union[str, Tuple[Any, str]]] = []

        for xmlsnippet in xmlsnippet.split(" "):
            pfx, tag = xmlsnippet.rsplit(":", 1)
            xmlns = self.prefixes.get(pfx, pfx)

            parts.append((xmlns, tag))

        if isentity:
            assert len(parts) == 1

            # mypy: Incompatible return value type (got "Union[str, Tuple[Any, str]]",
            # mypy: expected "Tuple[str, Union[str, Tuple[str, str]], str]")
            # franckly, I'm not sure here...
            return parts[0]  # type: ignore

        if len(parts) == 1:
            return "*", parts[0], "*"

        if len(parts) == 2:
            parts.append("*")

        assert len(parts) == 3

        # mypy: Incompatible return value type (got "Tuple[Union[str, Tuple[Any, str]], ...]",
        # mypy: expected "Tuple[str, Union[str, Tuple[str, str]], str]")
        # should be good because of the assert before
        return tuple(parts)  # type: ignore

    def add_equivalence(self, yamssnippet: str, xmlsnippet: str) -> None:
        ykey = self._norm_yams_key(yamssnippet)
        isentity = not isinstance(ykey, tuple)
        xkey = self._norm_xml_key(xmlsnippet, isentity=isentity)

        self._y2x.setdefault(ykey, []).append(xkey)
        self._x2y.setdefault(xkey, []).append(ykey)

        if not isentity:
            for dict_, key, value in ((self._x2y, xkey, ykey), (self._y2x, ykey, xkey)):
                if key[0] != "*":
                    wkey = ("*",) + key[1:]
                    dict_.setdefault(wkey, []).append(value)

                if key[2] != "*":
                    wkey = key[:2] + ("*",)
                    dict_.setdefault(wkey, []).append(value)

    def remove_equivalence(self, yamssnippet: str, xmlsnippet: str) -> None:
        ykey = self._norm_yams_key(yamssnippet)
        isentity = not isinstance(ykey, tuple)
        xkey = self._norm_xml_key(xmlsnippet, isentity=isentity)

        self._y2x.setdefault(ykey, []).remove(xkey)
        self._x2y.setdefault(xkey, []).remove(ykey)

        if not isentity:
            for dict_, key, value in ((self._x2y, xkey, ykey), (self._y2x, ykey, xkey)):
                if key[0] != "*":
                    wkey = ("*",) + key[1:]
                    dict_.setdefault(wkey, []).remove(value)

                if key[2] != "*":
                    wkey = key[:2] + ("*",)
                    dict_.setdefault(wkey, []).remove(value)

    def yeq(self, xmlsnippet: str, isentity: bool = False) -> List[XYSet]:
        key = self._norm_xml_key(xmlsnippet, isentity)
        try:
            return self._x2y[key]
        except KeyError:
            raise UnsupportedVocabulary(key)

    def xeq(self, yamssnippet: str) -> List[XYSet]:
        key = self._norm_yams_key(yamssnippet)
        try:
            return self._y2x[key]
        except KeyError:
            raise UnsupportedVocabulary(key)


XY: XYRegistry = XYRegistry()
register_prefix = XY.register_prefix
add_equivalence = XY.add_equivalence
remove_equivalence = XY.remove_equivalence
yeq = XY.yeq
xeq = XY.xeq



# copyright 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.

from yams.buildobjs import EntityType, RelationType, RelationDefinition, SubjectRelation, String


_ = str


class Company(EntityType):
    name = String()


class Subcompany(Company):
    __specializes_schema__ = True
    subcompany_of = SubjectRelation("Company")


class Division(Company):
    __specializes_schema__ = True
    division_of = SubjectRelation("Company")


class Subdivision(Division):
    __specializes_schema__ = True
    subdivision_of = SubjectRelation("Company")


class Employee(EntityType):
    works_for = SubjectRelation("Company")


class require_permission(RelationType):
    """link a permission to the entity. This permission should be used in the
    security definition of the entity's type to be useful.
    """

    fulltext_container = "subject"
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers",),
        "delete": ("managers",),
    }


class missing_require_permission(RelationDefinition):
    name = "require_permission"
    subject = "Company"
    object = "EPermission"


class EPermission(EntityType):
    """entity type that may be used to construct some advanced security configuration"""

    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": ("managers",),
        "delete": ("managers",),
        "update": (
            "managers",
            "owners",
        ),
    }
    name = String(
        required=True,
        indexed=True,
        internationalizable=True,
        fulltextindexed=True,
        maxsize=100,
        description=_("name or identifier of the permission"),
    )



# copyright 2004-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with yams. If not, see <https://www.gnu.org/licenses/>.
"""Object model and utilities to define generic Entities/Relations schemas.
"""
import warnings
from datetime import datetime, date

from typing import Set, Type, Dict, Union, Callable, Any, Iterable, TypeVar

import pkg_resources

from logilab.common.date import strptime_time
from logilab.common import nullobject

from yams._exceptions import SchemaError, UnknownType, BadSchemaDefinition, ValidationError  # noqa
import yams.types as yams_types


__docformat__: str = "restructuredtext en"

__version__: str = pkg_resources.get_distribution("yams").version

_: Type[str] = str

MARKER: nullobject = nullobject()

BASE_TYPES: Set[str] = set(
    (
        "String",
        "Password",
        "Bytes",
        "Int",
        "BigInt",
        "Float",
        "Boolean",
        "Decimal",
        "Date",
        "Time",
        "Datetime",
        "TZTime",
        "TZDatetime",
        "Interval",
    )
)

# base groups used in permissions
BASE_GROUPS: Set[str] = set((_("managers"), _("users"), _("guests"), _("owners")))

# default permissions for entity types, relations and attributes
DEFAULT_ETYPEPERMS: yams_types.Permissions = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "update": (
        "managers",
        "owners",
    ),
    "delete": ("managers", "owners"),
    "add": (
        "managers",
        "users",
    ),
}
DEFAULT_RELPERMS: yams_types.Permissions = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "delete": ("managers", "users"),
    "add": (
        "managers",
        "users",
    ),
}
DEFAULT_ATTRPERMS: yams_types.Permissions = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "add": ("managers", "users"),
    "update": ("managers", "owners"),
}
DEFAULT_COMPUTED_RELPERMS: yams_types.Permissions = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "delete": (),
    "add": (),
}
DEFAULT_COMPUTED_ATTRPERMS: yams_types.Permissions = {
    "read": (
        "managers",
        "users",
        "guests",
    ),
    "add": (),
    "update": (),
}

# This provides a way to specify callable objects as default values
# First level is the final type, second is the keyword to callable map
_current_date_or_datetime_constructor_type = Callable[[], Union[date, datetime]]
KEYWORD_MAP: Dict[str, Dict[str, _current_date_or_datetime_constructor_type]] = {
    "Datetime": {"NOW": datetime.now, "TODAY": datetime.today},
    "TZDatetime": {"NOW": datetime.utcnow, "TODAY": datetime.today},
    "Date": {"TODAY": date.today},
}

# bw compat for literal date/time values stored as strings in schemas
DATE_FACTORY_MAP: Dict[str, Callable[[str], Union[datetime, float]]] = {
    "Datetime": lambda x: ":" in x
    and datetime.strptime(x, "%Y/%m/%d %H:%M")
    or datetime.strptime(x, "%Y/%m/%d"),
    "Date": lambda x: datetime.strptime(x, "%Y/%m/%d"),
    "Time": strptime_time,  # returns a float (from time())
}

KNOWN_METAATTRIBUTES: Set[str] = set(("format", "encoding", "name"))


_type_of_default = TypeVar("_type_of_default")


def convert_default_value(
    rdef: yams_types.RdefRdefSchema, default: _type_of_default
) -> Union[_type_of_default, datetime, date, float, str, int, bool]:
    # rdef can be either a .schema.RelationDefinitionSchema
    # or a .buildobjs.RelationDefinition

    # mypy ouput: Item "RelationDefinition" of "Union[RelationDefinitionSchema,
    # mypy output: RelationDefinition]" has no attribute "relation_type"
    # RelationDefinition will succed on having a name so it won't jump to ".relation_type"
    relation_type = getattr(rdef, "name", None) or rdef.relation_type.type  # type: ignore

    if isinstance(default, str) and rdef.object != "String":
        # real Strings can be anything,
        # including things that look like keywords for other base types
        if str(rdef.object) in KEYWORD_MAP and default.upper() in KEYWORD_MAP[str(rdef.object)]:
            return KEYWORD_MAP[str(rdef.object)][default.upper()]()
        # else:
        # the default was likely not a special constant
        # like TODAY but some literal

        # bw compat for old schemas
        if rdef.object in DATE_FACTORY_MAP:
            warnings.warn(
                "using strings as default values "
                "for attribute %s of type %s "
                "is deprecated; you should use "
                "the plain python objects instead" % (relation_type, rdef.object),
                DeprecationWarning,
            )

            try:
                return DATE_FACTORY_MAP[str(rdef.object)](default)
            except ValueError as verr:
                raise ValueError(
                    "creating a default value for "
                    "attribute %s of type %s from the string %r"
                    " is not supported (cause %s)" % (relation_type, rdef.object, default, verr)
                )

    if rdef.object == "String":
        return str(default)

    return default  # general case: untouched default


def register_base_type(
    name: str,
    parameters: Union[Dict[str, Any], Iterable[str]] = (),
    check_function: Callable = None,
) -> None:
    """register a yams base (final) type. You'll have to call
    base_type_class to generate the class.
    """
    from yams.schema import RelationDefinitionSchema
    from yams.constraints import BASE_CHECKERS, yes

    # Add the datatype to yams base types
    assert name not in BASE_TYPES, "%s already in BASE_TYPES %s" % (name, BASE_TYPES)

    BASE_TYPES.add(name)

    # Add the new datatype to the authorized types of RelationDefinitionSchema
    if not isinstance(parameters, dict):
        # turn tuple/list into dict with None values
        parameters = dict((p, None) for p in parameters)

    RelationDefinitionSchema.BASE_TYPE_PROPERTIES[name] = parameters

    # Add a yams checker or yes is not specified
    BASE_CHECKERS[name] = check_function or yes


def unregister_base_type(name: str) -> None:
    """Unregister a yams base (final) type"""
    from yams.schema import RelationDefinitionSchema
    from yams.constraints import BASE_CHECKERS

    assert name in BASE_TYPES, "%s not in BASE_TYPES %s" % (name, BASE_TYPES)

    BASE_TYPES.remove(name)
    RelationDefinitionSchema.BASE_TYPE_PROPERTIES.pop(name)
    BASE_CHECKERS.pop(name)



# copyright 2004-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with yams. If not, see <https://www.gnu.org/licenses/>.
"""Some common constraint classes."""

import re
import decimal
import operator
import json
import datetime
import warnings

from typing import Any, Dict, Match, Tuple, Type, Union, Optional, Callable, Sequence, List, cast

from logilab.common.deprecation import class_renamed

import yams
from yams import BadSchemaDefinition
from yams.interfaces import IConstraint, IVocabularyConstraint
import yams.types as yams_types

__docformat__: str = "restructuredtext en"

_: Type[str] = str


class ConstraintJSONEncoder(json.JSONEncoder):
    def default(self, obj: Union[Any, "NOW", "TODAY"]) -> Union[Any, dict]:
        if isinstance(obj, Attribute):
            return {"__attribute__": obj.attr}

        if isinstance(obj, NOW):
            # it is not a timedelta
            if obj.offset is None:
                return {"__now__": True, "offset": obj.offset}

            d = {
                "days": obj.offset.days,
                "seconds": obj.offset.seconds,
                "microseconds": obj.offset.microseconds,
            }

            return {"__now__": True, "offset": d}

        if isinstance(obj, TODAY):
            # it is not a timedelta
            if obj.offset is None:
                return {"__today__": True, "offset": obj.offset, "type": obj.type}

            d = {
                "days": obj.offset.days,
                "seconds": obj.offset.seconds,
                "microseconds": obj.offset.microseconds,
            }

            return {"__today__": True, "offset": d, "type": obj.type}

        return super(ConstraintJSONEncoder, self).default(obj)


def _json_object_hook(dct: Dict) -> Union[Dict, "NOW", "TODAY", "Attribute"]:
    offset: Optional[datetime.timedelta]

    if "__attribute__" in dct:
        return Attribute(dct["__attribute__"])

    if "__now__" in dct:
        if dct["offset"] is not None:
            offset = datetime.timedelta(**dct["offset"])
        else:
            offset = None

        return NOW(offset)

    if "__today__" in dct:
        if dct["offset"] is not None:
            offset = datetime.timedelta(**dct["offset"])
        else:
            offset = None

        return TODAY(offset=offset, type=dct["type"])

    return dct


def cstr_json_dumps(obj: yams_types.jsonSerializable) -> str:
    return str(ConstraintJSONEncoder(sort_keys=True).encode(obj))


cstr_json_loads: Callable[[str], Dict] = json.JSONDecoder(object_hook=_json_object_hook).decode


def _message_value(boundary) -> Any:
    if isinstance(boundary, Attribute):
        return boundary.attr
    return boundary


class BaseConstraint:
    """base class for constraints"""

    __implements__ = IConstraint

    def __init__(self, msg: Optional[str] = None) -> None:
        self.msg: Optional[str] = msg

    def check_consistency(
        self,
        subjschema: yams_types.EntitySchema,
        objschema: yams_types.EntitySchema,
        rdef: yams_types.RelationDefinition,
    ) -> None:
        pass

    def type(self) -> str:
        return self.__class__.__name__

    def serialize(self) -> str:
        """called to make persistent valuable data of a constraint"""
        return cstr_json_dumps({"msg": self.msg})

    @classmethod
    def deserialize(cls: Type["BaseConstraint"], value: str) -> Any:
        """called to restore serialized data of a constraint. Should return
        a `cls` instance
        """
        value = value.strip()

        if value and value != "None":
            d = cstr_json_loads(value)
        else:
            d = {}

        return cls(**d)

    def failed_message(self, key: str, value, entity=None) -> Tuple[Optional[str], Dict[str, Any]]:
        if entity is None:
            warnings.warn(
                "[yams 0.44] failed message " "should now be given entity has argument.",
                DeprecationWarning,
                stacklevel=2,
            )

        if self.msg:
            return self.msg, {}

        return self._failed_message(entity, key, value)

    def _failed_message(self, entity, key: str, value) -> Tuple[str, Dict[str, Any]]:
        return (
            _("%(KEY-cstr)s constraint failed for value %(KEY-value)r"),
            {key + "-cstr": self, key + "-value": value},
        )

    def __eq__(self, other: Any) -> bool:
        return (self.type(), self.serialize()) == (other.type(), other.serialize())

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __lt__(self, other: Any) -> bool:
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self.type(), self.serialize()))


# possible constraints ########################################################


class UniqueConstraint(BaseConstraint):
    """object of relation must be unique"""

    def __str__(self) -> str:
        return "unique"

    def check_consistency(
        self,
        subjschema: yams_types.EntitySchema,
        objschema: yams_types.EntitySchema,
        rdef: yams_types.RelationDefinition,
    ) -> None:

        if not objschema.final:
            raise BadSchemaDefinition("unique constraint doesn't apply to non " "final entity type")

    def check(self, entity, rtype: yams_types.RelationType, values) -> bool:
        """return true if the value satisfy the constraint, else false"""
        return True


class SizeConstraint(BaseConstraint):
    """the string size constraint :

    if max is not None the string length must not be greater than max
    if min is not None the string length must not be shorter than min
    """

    def __init__(
        self, max: Optional[int] = None, min: Optional[int] = None, msg: Optional[str] = None
    ) -> None:

        super(SizeConstraint, self).__init__(msg)

        assert max is not None or min is not None, "No max or min"

        if min is not None:
            assert isinstance(min, int), "min must be an int, not %r" % min

        if max is not None:
            assert isinstance(max, int), "max must be an int, not %r" % max

        self.max: Optional[int] = max
        self.min: Optional[int] = min

    def __str__(self) -> str:
        res = "size"

        if self.max is not None:
            res = "%s <= %s" % (res, self.max)

        if self.min is not None:
            res = "%s <= %s" % (self.min, res)

        return res

    def check_consistency(
        self,
        subjschema: yams_types.EntitySchema,
        objschema: yams_types.EntitySchema,
        rdef: yams_types.RelationDefinition,
    ) -> None:

        if not objschema.final:
            raise BadSchemaDefinition("size constraint doesn't apply to non " "final entity type")

        if objschema not in ("String", "Bytes", "Password"):
            raise BadSchemaDefinition(
                "size constraint doesn't apply to %s " "entity type" % objschema
            )

        if self.max:
            for cstr in rdef.constraints:
                if cstr.__class__ is StaticVocabularyConstraint:
                    for value in cstr.values:
                        if len(value) > self.max:
                            raise BadSchemaDefinition(
                                "size constraint set to %s but vocabulary "
                                "contains string of greater size" % self.max
                            )

    def check(self, entity, rtype: yams_types.RelationType, value: Sequence) -> bool:
        """return true if the value is in the interval specified by
        self.min and self.max
        """
        if self.max is not None and len(value) > self.max:
            return False

        if self.min is not None and len(value) < self.min:
            return False

        return True

    def _failed_message(self, entity, key: str, value: Sequence) -> Tuple[str, Dict[str, Any]]:
        if self.max is not None and len(value) > self.max:
            return (
                _("value should have maximum size of %(KEY-max)s" " but found %(KEY-size)s"),
                {key + "-max": self.max, key + "-size": len(value)},
            )

        if self.min is not None and len(value) < self.min:
            return (
                _("value should have minimum size of %(KEY-min)s" " but found %(KEY-size)s"),
                {key + "-min": self.min, key + "-size": len(value)},
            )

        assert False, "shouldnt be there"

    def serialize(self) -> str:
        """simple text serialization"""
        return cstr_json_dumps({"min": self.min, "max": self.max, "msg": self.msg})

    @classmethod
    def deserialize(cls: Type["SizeConstraint"], value: str) -> "SizeConstraint":
        """simple text deserialization"""
        try:
            d = cstr_json_loads(value)

            return cls(**d)
        except ValueError:
            kwargs = {}

            for adef in value.split(","):
                key, val = [w.strip() for w in adef.split("=")]

                assert key in ("min", "max")

                kwargs[str(key)] = int(val)

            # mypy: Argument 1 to "SizeConstraint" has incompatible type "**Dict[str, int]";
            # mypy: expected "Optional[str]"
            # mypy seems really broken with **kwargs
            return cls(**kwargs)  # type: ignore


class RegexpConstraint(BaseConstraint):
    """specifies a set of allowed patterns for a string value"""

    __implements__ = IConstraint

    def __init__(self, regexp: str, flags: int = 0, msg: Optional[str] = None) -> None:
        """
        Construct a new RegexpConstraint.

        :Parameters:
         - `regexp`: (str) regular expression that strings must match
         - `flags`: (int) flags that are passed to re.compile()
        """
        super(RegexpConstraint, self).__init__(msg)
        self.regexp: str = regexp
        self.flags: int = flags
        self._rgx = re.compile(regexp, flags)

    def __str__(self) -> str:
        return "regexp %s" % self.regexp

    def check_consistency(
        self,
        subjschema: yams_types.EntitySchema,
        objschema: yams_types.EntitySchema,
        rdef: yams_types.RelationDefinition,
    ) -> None:
        if not objschema.final:
            raise BadSchemaDefinition("regexp constraint doesn't apply to non " "final entity type")

        if objschema not in ("String", "Password"):
            raise BadSchemaDefinition(
                "regexp constraint doesn't apply to %s " "entity type" % objschema
            )

    def check(self, entity, rtype: yams_types.RelationType, value: str) -> Optional[Match[str]]:
        """return true if the value maches the regular expression"""
        return self._rgx.match(value, self.flags)

    def _failed_message(self, entity, key: str, value) -> Tuple[str, Dict[str, Any]]:
        return (
            _("%(KEY-value)r doesn't match " "the %(KEY-regexp)r regular expression"),
            {key + "-value": value, key + "-regexp": self.regexp},
        )

    def serialize(self) -> str:
        """simple text serialization"""
        return cstr_json_dumps({"regexp": self.regexp, "flags": self.flags, "msg": self.msg})

    @classmethod
    def deserialize(cls, value: str) -> "RegexpConstraint":
        """simple text deserialization"""
        try:
            d = cstr_json_loads(value)
            return cls(**d)
        except ValueError:
            regexp, flags = value.rsplit(",", 1)
            return cls(regexp, int(flags))

    def __deepcopy__(self, memo) -> "RegexpConstraint":
        return RegexpConstraint(self.regexp, self.flags)


OPERATORS: Dict[str, Callable[[Any, Any], bool]] = {
    "<=": operator.le,
    "<": operator.lt,
    ">": operator.gt,
    ">=": operator.ge,
}


class BoundaryConstraint(BaseConstraint):
    """the int/float bound constraint :

    set a minimal or maximal value to a numerical value
    """

    __implements__ = IConstraint

    def __init__(
        self, op: str, boundary: Optional[Union["Attribute", "NOW", "TODAY"]] = None, msg=None
    ) -> None:

        super(BoundaryConstraint, self).__init__(msg)

        assert op in OPERATORS, op

        self.operator: str = op
        self.boundary: Optional[Union["Attribute", "NOW", "TODAY"]] = boundary

    def __str__(self) -> str:
        return "value %s" % self.serialize()

    def check_consistency(
        self,
        subjschema: yams_types.EntitySchema,
        objschema: yams_types.EntitySchema,
        rdef: yams_types.RelationDefinition,
    ) -> None:

        if not objschema.final:
            raise BadSchemaDefinition("bound constraint doesn't apply to non " "final entity type")

    def check(self, entity, rtype: yams_types.RelationType, value) -> bool:
        """return true if the value satisfies the constraint, else false"""
        boundary = actual_value(self.boundary, entity)

        if boundary is None:
            return True

        return OPERATORS[self.operator](value, boundary)

    def _failed_message(self, entity, key: str, value) -> Tuple[str, Dict[str, Any]]:
        return (
            "value %%(KEY-value)s must be %s %%(KEY-boundary)s" % self.operator,
            {
                key + "-value": value,
                key + "-boundary": _message_value(actual_value(self.boundary, entity)),
            },
        )

    def serialize(self) -> str:
        """simple text serialization"""
        return cstr_json_dumps({"op": self.operator, "boundary": self.boundary, "msg": self.msg})

    @classmethod
    def deserialize(cls: Type["BoundaryConstraint"], value: str) -> "BoundaryConstraint":
        """simple text deserialization"""
        try:
            d = cstr_json_loads(value)

            return cls(**d)
        except ValueError:
            op, boundary = value.split(" ", 1)

            return cls(op, eval(boundary))


BoundConstraint = class_renamed("BoundConstraint", BoundaryConstraint)
BoundConstraint.type = lambda x: "BoundaryConstraint"

_("value %(KEY-value)s must be < %(KEY-boundary)s")
_("value %(KEY-value)s must be > %(KEY-boundary)s")
_("value %(KEY-value)s must be <= %(KEY-boundary)s")
_("value %(KEY-value)s must be >= %(KEY-boundary)s")


class IntervalBoundConstraint(BaseConstraint):
    """an int/float bound constraint :

    sets a minimal and / or a maximal value to a numerical value
    This class replaces the BoundConstraint class
    """

    __implements__ = IConstraint

    def __init__(
        self,
        minvalue: Optional[Union[int, float]] = None,
        maxvalue: Optional[Union[int, float]] = None,
        msg: Optional[str] = None,
    ) -> None:
        """
        :param minvalue: the minimal value that can be used
        :param maxvalue: the maxvalue value that can be used
        """
        assert not (minvalue is None and maxvalue is None)

        super(IntervalBoundConstraint, self).__init__(msg)

        self.minvalue: Optional[Union[int, float]] = minvalue
        self.maxvalue: Optional[Union[int, float]] = maxvalue

    def __str__(self) -> str:
        return "value [%s]" % self.serialize()

    def check_consistency(
        self,
        subjschema: yams_types.EntitySchema,
        objschema: yams_types.EntitySchema,
        rdef: yams_types.RelationDefinition,
    ) -> None:

        if not objschema.final:
            raise BadSchemaDefinition(
                "interval bound constraint doesn't apply" " to non final entity type"
            )

    def check(self, entity, rtype: yams_types.RelationType, value: Union[int, float]) -> bool:
        minvalue = actual_value(self.minvalue, entity)

        if minvalue is not None and value < minvalue:
            return False

        maxvalue = actual_value(self.maxvalue, entity)

        if maxvalue is not None and value > maxvalue:
            return False

        return True

    def _failed_message(self, entity, key: str, value) -> Tuple[str, Dict[str, Any]]:
        if self.minvalue is not None and value < actual_value(self.minvalue, entity):
            return (
                _("value %(KEY-value)s must be >= %(KEY-boundary)s"),
                {key + "-value": value, key + "-boundary": _message_value(self.minvalue)},
            )

        if self.maxvalue is not None and value > actual_value(self.maxvalue, entity):
            return (
                _("value %(KEY-value)s must be <= %(KEY-boundary)s"),
                {key + "-value": value, key + "-boundary": _message_value(self.maxvalue)},
            )

        assert False, "shouldnt be there"

    def serialize(self) -> str:
        """simple text serialization"""
        return cstr_json_dumps(
            {"minvalue": self.minvalue, "maxvalue": self.maxvalue, "msg": self.msg}
        )

    @classmethod
    def deserialize(cls: Type["IntervalBoundConstraint"], value: str) -> "IntervalBoundConstraint":
        """simple text deserialization"""
        try:
            d = cstr_json_loads(value)

            return cls(**d)
        except ValueError:
            minvalue, maxvalue = value.split(";")

            return cls(eval(minvalue), eval(maxvalue))


class StaticVocabularyConstraint(BaseConstraint):
    """Enforces a predefined vocabulary set for the value."""

    __implements__ = IVocabularyConstraint

    def __init__(self, values: Sequence[str], msg: Optional[str] = None) -> None:
        super(StaticVocabularyConstraint, self).__init__(msg)
        self.values: Tuple[str, ...] = tuple(values)

    def __str__(self) -> str:
        return "value in (%s)" % ", ".join(repr(str(word)) for word in self.vocabulary())

    def check(self, entity, rtype: yams_types.RelationType, value: str) -> bool:
        """return true if the value is in the specific vocabulary"""
        return value in self.vocabulary(entity=entity)

    def _failed_message(self, entity, key: str, value) -> Tuple[str, Dict[str, Any]]:
        if isinstance(value, str):
            value = '"%s"' % str(value)
            choices = ", ".join('"%s"' % val for val in self.values)
        else:
            choices = ", ".join(str(val) for val in self.values)

        return (
            _("invalid value %(KEY-value)s, " "it must be one of %(KEY-choices)s"),
            {key + "-value": value, key + "-choices": choices},
        )

    def vocabulary(self, **kwargs) -> Tuple[str, ...]:
        """return a list of possible values for the attribute"""
        return self.values

    def serialize(self) -> str:
        """serialize possible values as a json object"""
        return cstr_json_dumps({"values": self.values, "msg": self.msg})

    @classmethod
    def deserialize(
        cls: Type["StaticVocabularyConstraint"], value: str
    ) -> "StaticVocabularyConstraint":
        """deserialize possible values from a csv list of evaluable strings"""
        try:
            values = cstr_json_loads(value)

            return cls(**values)
        except ValueError:
            interpreted_values = [eval(w) for w in re.split("(?<!,), ", value)]

            if interpreted_values and isinstance(interpreted_values[0], str):
                cast(List[str], interpreted_values)

                interpreted_values = [v.replace(",,", ",") for v in interpreted_values]

            return cls(interpreted_values)


class FormatConstraint(StaticVocabularyConstraint):

    regular_formats: Tuple[str, ...] = (
        _("text/rest"),
        _("text/markdown"),
        _("text/html"),
        _("text/plain"),
    )

    # **kwargs to have a common interface between all Constraint initializers
    def __init__(self, msg: Optional[str] = None, **kwargs) -> None:
        values: Tuple[str, ...] = self.regular_formats
        super(FormatConstraint, self).__init__(values, msg=msg)

    def check_consistency(
        self,
        subjschema: yams_types.EntitySchema,
        objschema: yams_types.EntitySchema,
        rdef: yams_types.RelationDefinition,
    ) -> None:

        if not objschema.final:
            raise BadSchemaDefinition("format constraint doesn't apply to non " "final entity type")

        if not objschema == "String":
            raise BadSchemaDefinition("format constraint only apply to String")


FORMAT_CONSTRAINT: FormatConstraint = FormatConstraint()


class MultipleStaticVocabularyConstraint(StaticVocabularyConstraint):
    """Enforce a list of values to be in a predefined set vocabulary."""

    # XXX never used

    def check(self, entity, rtype: yams_types.RelationType, values: Sequence[str]) -> bool:
        """return true if the values satisfy the constraint, else false"""
        vocab = self.vocabulary(entity=entity)

        for value in values:
            if value not in vocab:
                return False

        return True


# special classes to be used w/ constraints accepting values as argument(s):
# IntervalBoundConstraint


def actual_value(value, entity) -> Any:
    if hasattr(value, "value"):
        return value.value(entity)

    return value


class Attribute:
    def __init__(self, attr) -> None:
        self.attr = attr

    def __str__(self) -> str:
        return "%s(%r)" % (self.__class__.__name__, self.attr)

    def value(self, entity) -> Any:
        return getattr(entity, self.attr)


class NOW:
    def __init__(self, offset: Optional[datetime.timedelta] = None) -> None:
        self.offset: Optional[datetime.timedelta] = offset

    def __str__(self) -> str:
        return "%s(%r)" % (self.__class__.__name__, self.offset)

    def value(self, entity) -> datetime.date:
        now = yams.KEYWORD_MAP["Datetime"]["NOW"]()

        if self.offset:
            now += self.offset

        return now


class TODAY:
    def __init__(self, offset: Optional[datetime.timedelta] = None, type: str = "Date") -> None:
        self.offset: Optional[datetime.timedelta] = offset
        # XXX no check that self.type is in KEYWORD_MAP?
        self.type: str = type

    def __str__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.offset, self.type)

    def value(self, entity) -> datetime.date:
        now = yams.KEYWORD_MAP[self.type]["TODAY"]()

        if self.offset:
            now += self.offset

        return now


# base types checking functions ###############################################


def check_string(eschema, value) -> bool:
    """check value is an unicode string"""
    return isinstance(value, str)


def check_password(eschema, value) -> bool:
    """check value is an encoded string"""
    return isinstance(value, bytes)


def check_int(eschema, value) -> bool:
    """check value is an integer"""
    try:
        int(value)
    except ValueError:
        return False

    return True


def check_float(eschema, value) -> bool:
    """check value is a float"""
    try:
        float(value)
    except ValueError:
        return False

    return True


def check_decimal(eschema, value) -> bool:
    """check value is a Decimal"""
    try:
        decimal.Decimal(value)
    except (TypeError, decimal.InvalidOperation):
        return False

    return True


def check_boolean(eschema, value) -> bool:
    """check value is a boolean"""
    return isinstance(value, int)


def check_file(eschema, value) -> bool:
    """check value has a getvalue() method (e.g. StringIO or cStringIO)"""
    return hasattr(value, "getvalue")


def yes(*args, **kwargs) -> bool:
    """dunno how to check"""
    return True


BASE_CHECKERS: yams_types.Checkers = {
    "Date": yes,
    "Time": yes,
    "Datetime": yes,
    "TZTime": yes,
    "TZDatetime": yes,
    "Interval": yes,
    "String": check_string,
    "Int": check_int,
    "BigInt": check_int,
    "Float": check_float,
    "Decimal": check_decimal,
    "Boolean": check_boolean,
    "Password": check_password,
    "Bytes": check_file,
}

BASE_CONVERTERS: yams_types.Converters = {
    "String": str,
    "Password": bytes,
    "Int": int,
    "BigInt": int,
    "Float": float,
    "Boolean": bool,
    "Decimal": decimal.Decimal,
}


def patch_sqlite_decimal() -> None:
    """patch Decimal checker and converter to bypass SQLITE Bug
    (SUM of Decimal return float in SQLITE)"""

    def convert_decimal(value) -> decimal.Decimal:
        # XXX issue a warning
        if isinstance(value, float):
            value = str(value)

        return decimal.Decimal(value)

    def check_decimal(eschema, value) -> bool:
        """check value is a Decimal"""
        try:
            if isinstance(value, float):
                return True

            decimal.Decimal(value)
        except (TypeError, decimal.InvalidOperation):
            return False

        return True

    global BASE_CONVERTERS
    BASE_CONVERTERS["Decimal"] = convert_decimal

    global BASE_CHECKERS
    BASE_CHECKERS["Decimal"] = check_decimal
